from hdlConvertorAst.hdlAst import HdlOp, HdlStmIf, HdlStmWhile, HdlStmBlock, HdlStmAssign, HdlStmCaseType
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last
from hdlConvertorAst.to.hwt.expr import ToHwtExpr
from hdlConvertorAst.to.basic_hdl_sim_model import ToBasicHdlSimModel


class ToHwtStm(ToHwtExpr):

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        w = self.out.write
        if o.labels:
            w("# ")
            if o.labels:
                w(o.labels[0])
                # w(", ")
        # w("sens: ")
        # if o.sensitivity:
        #    for last, s in iter_with_last(o.sensitivity):
        #        if isinstance(s, HdlOp):
        #            w(str(s.fn))
        #            w(" ")
        #            self.visit_iHdlExpr(s.ops[0])
        #        else:
        #            self.visit_iHdlExpr(s)
        #        if not last:
        #            w(", ")
        w("\n")
        self.visit_doc(o)
        if o.trigger_constrain is not None:
            raise NotImplementedError()
        self.visit_iHdlStatement(o.body)
        # w("\n")

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        self.visit_doc(o)
        w = self.out.write
        for is_last, i in iter_with_last(o.body):
            self.visit_iHdlStatement(i)
            if not is_last:
                if o.in_preproc:
                    w("\n")
                else:
                    w(",\n")

    def visit_HdlStmIf(self, o):
        """
        :type stm: HdlStmIf
        """
        self.visit_doc(o)
        w = self.out.write

        in_preproc = o.in_preproc
        if in_preproc:
            w("if ")
        else:
            w("If(")
        self.visit_iHdlExpr(o.cond)
        if in_preproc:
            w(":\n")
        else:
            w(",\n")
        with Indent(self.out):
            self.visit_iHdlStatement(o.if_true)
            w("\n")
        if not in_preproc:
            w(")")
        for (c, _stm) in o.elifs:
            if in_preproc:
                w("elif ")
            else:
                w(".Elif(")
            self.visit_iHdlExpr(c)
            if in_preproc:
                w(":\n")
            else:
                w(",\n")
            with Indent(self.out):
                self.visit_iHdlStatement(_stm)
            w("\n")
            if not in_preproc:
                w(")")

        ifFalse = o.if_false
        if ifFalse is not None:
            if in_preproc:
                w("else:\n")
            else:
                w(".Else(\n")
            with Indent(self.out):
                self.visit_iHdlStatement(ifFalse)
                w("\n")
            if not in_preproc:
                w(")")

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        self.visit_doc(o)
        w = self.out.write
        self.visit_iHdlExpr(o.dst)
        if o.is_blocking:
            raise NotImplementedError(o)
        if o.time_delay is not None:
            raise NotImplementedError()
        if o.event_delay is not None:
            raise NotImplementedError()
        if o.in_preproc:
            w(" = ")
            self.visit_iHdlExpr(o.src)
        else:
            w("(")
            self.visit_iHdlExpr(o.src)
            w(")")

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        self.visit_doc(o)
        if o.uniq_constrain is not None:
            raise NotImplementedError()
        w = self.out.write
        if o.in_preproc:
            assert o.cases
            first = True
            for c, stm in o.cases:
                if first:
                    w("if ")
                    first = False
                else:
                    w("elif ")
                self.visit_iHdlExpr(o.switch_on)
                w(" == ")
                self.visit_iHdlExpr(c)
                w(":\n")
                with Indent(self.out):
                    self.visit_iHdlStatement(stm)
                w("\n")

            if o.default is not None:
                w("else:\n")
                with Indent(self.out):
                    self.visit_iHdlStatement(o.default)

        else:
            # if o.type != HdlStmCaseType.CASE:
            #    raise NotImplementedError(o.type)
            w("Switch(")
            self.visit_iHdlExpr(o.switch_on)
            w(")")
            with Indent(self.out):
                for c, stm in o.cases:
                    w("\\\n")
                    w(".Case(")
                    self.visit_iHdlExpr(c)
                    w(",\n")
                    with Indent(self.out):
                        self.visit_iHdlStatement(stm)
                    w(")")
                if o.default is not None:
                    w("\\\n")
                    w(".Default(\n")
                    with Indent(self.out):
                        self.visit_iHdlStatement(o.default)
                        w(")")

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        if not o.in_preproc:
            raise TypeError("does not support HdlStmFor", self, o)
        self.visit_doc(o)
        w = self.out.write
        # [todo] if is a simple for in range()

        self.visit_iHdlObj(o.init)
        w("\n")
        w("while ")
        self.visit_iHdlExpr(o.cond)
        w(":\n")
        with Indent(self.out):
            self.visit_iHdlObj(o.body)

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        """
        self.visit_doc(o)
        assert o.in_preproc
        w = self.out.write
        w("while ")
        self.visit_iHdlExpr(o.cond)
        w(":\n")
        with Indent(self.out):
            self.visit_iHdlObj(o.body)

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        assert o.in_preproc, o
        self.out.write("break")
    
    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        assert o.in_preproc, o
        self.out.write("continue")
    
    def visit_HdlStmThrow(self, o):
        ToBasicHdlSimModel.visit_HdlStmThrow(self, o)

    def visit_HdlStmWait(self, o):
        ToBasicHdlSimModel.visit_HdlStmWait(self, o)

    def visit_HdlStmNop(self, o):
        ToBasicHdlSimModel.visit_HdlStmNop(self, o)



"""
This module contains a convertor from HdlConvertor AST
to a BasicHldSimModel code.

BasicHldSimModel is a simple HDL like python component description
which can be directly simulated by Discrete-event simulators like
https://github.com/Nic30/hwtSimApi.git
"""

from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel



from enum import Enum

from hdlConvertorAst.hdlAst import HdlModuleDec, HdlOp
from hdlConvertorAst.hdlAst._expr import HdlOpType, HdlValueId, HdlValueInt
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.hdlUtils import AutoIndentingStream, iter_with_last
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor


# https://www.geeksforgeeks.org/operator-precedence-and-associativity-in-c/
# http://www.euroelectronica.ro/7-operators/
# https://gist.github.com/kputnam/5625856
# https://github.com/kaitai-io/kaitai_struct/issues/69
class ASSOCIATIVITY(Enum):
    L_TO_R = "L_TO_R"
    R_TO_L = "R_TO_L"
    NONE = "NONE"


ASSIGN_OPERATORS_SYMBOLS_C = {
    HdlOpType.ASSIGN: ' = ',
    HdlOpType.PLUS_ASSIGN: ' += ',
    HdlOpType.MINUS_ASSIGN: ' -= ',
    HdlOpType.MUL_ASSIGN: ' *= ',
    HdlOpType.DIV_ASSIGN: ' /= ',
    HdlOpType.MOD_ASSIGN: ' %= ',
    HdlOpType.AND_ASSIGN: ' &= ',
    HdlOpType.OR_ASSIGN: ' |= ',
    HdlOpType.XOR_ASSIGN: ' ^= ',
    HdlOpType.SHIFT_LEFT_ASSIGN: ' <<= ',
    HdlOpType.SHIFT_RIGHT_ASSIGN: ' >>= ',
}


class ToHdlCommon(HdlAstVisitor):

    INDENT_STEP = "    "
    ALL_UNARY_OPS = {
        getattr(HdlOpType, name) for name in dir(HdlOpType)
        if name.endswith("_UNARY")
    }
    GENERIC_UNARY_OPS = {
        HdlOpType.PLUS_UNARY: "+",
        HdlOpType.MINUS_UNARY: "-",
    }
    GENERIC_UNARY_OPS_POSTFIX = {}
    GENERIC_BIN_OPS = {
        HdlOpType.ADD: " + ",
        HdlOpType.SUB: " - ",
        HdlOpType.MUL: " * ",

        HdlOpType.LT: " < ",
        HdlOpType.LE: " <= ",
        HdlOpType.GT: " > ",
        HdlOpType.GE: " >= ",

        HdlOpType.DOT: ".",
    }

    def __init__(self, out_stream):
        super(ToHdlCommon, self).__init__()
        self.out = AutoIndentingStream(out_stream, self.INDENT_STEP)

    def visit_doc(self, obj, line_comment_prefix):
        """
        Format doc as line comments

        :type line_comment_prefix: str
        """
        doc = obj.doc
        if doc is not None:
            doc = doc.split("\n")
            w = self.out.write
            for last, d in iter_with_last(doc):
                if last and d == "":
                    break
                w(line_comment_prefix)
                w(d.replace('\r', ''))
                w("\n")

    def _precedence_of_expr(self, o):
        """
        :type o: iHdlExpr
        """
        # not id or value
        if not isinstance(o, HdlOp):
            return (-1, ASSOCIATIVITY.NONE, None)
        return self.OP_PRECEDENCE[o.fn] + (o.fn,)

    def visit_HdlOp(self, op):
        """
        :type op: HdlOp
        """
        o = op.fn
        w = self.out.write
        argc = len(op.ops)
        if argc == 1:
            op_str = self.GENERIC_UNARY_OPS.get(o, None)
            if op_str is not None:
                w(op_str)
                self._visit_operand(op.ops[0], 0, op, False, False)
                return
            op_str = self.GENERIC_UNARY_OPS_POSTFIX.get(o, None)
            if op_str is not None:
                self._visit_operand(op.ops[0], 0, op, False, False)
                w(op_str)
                return

        if argc == 2:
            op_str = self.GENERIC_BIN_OPS.get(o, None)
            if op_str is not None:
                return self._visit_bin_op(op, op_str)
        if o == HdlOpType.INDEX:
            return self._visit_operator_index(op)
        elif o == HdlOpType.CALL or o == HdlOpType.PARAMETRIZATION:
            return self.visit_operator_call(op)
        else:
            raise NotImplementedError(
                "Do not know how to convert %s argc:%d" % (o, argc))

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        """
        w = self.out.write
        if isinstance(o, HdlValueId):
            w(o.val)
            return
        elif is_str(o):
            w('"%s"' % o.replace("\n", "\\\n"))
            return
        elif isinstance(o, HdlValueInt):
            self.visit_HdlValueInt(o)
            return
        elif isinstance(o, HdlOp):
            self.visit_HdlOp(o)
            return
        elif isinstance(o, float):
            w("%e" % o)
        else:
            raise NotImplementedError(
                "Do not know how to convert %r" % (o))

    def _visit_operand_parentheses_extra_check(
            self,
            op_my, precedence_my, asoc_my,
            op_parent, precedence_parent, asoc_parent,
            left, right):
        if op_my in self.ALL_UNARY_OPS and op_parent in self.ALL_UNARY_OPS:
            return True
        return False

    def _visit_operand(self, operand, i,
                       parent,
                       expr_requires_parenthesis,
                       cancel_parenthesis):
        """
        :type operand: iHdlExpr
        :type i: int
        :type parent: HdlOp
        :type expr_requires_parenthesis: bool
        :type cancel_parenthesis: bool
        """
        use_parenthesis = False
        if not cancel_parenthesis:
            # resolve if the parenthesis are required
            precedence_my, asoc_my, op_my = self._precedence_of_expr(operand)
            if precedence_my >= 0:  # if this is an expression
                if expr_requires_parenthesis or asoc_my is ASSOCIATIVITY.NONE:
                    use_parenthesis = True
                else:
                    precedence_parent, asoc_parent = self.OP_PRECEDENCE[parent.fn]
                    right = None
                    left = None
                    argc = len(parent.ops)
                    assert argc, parent
                    if argc == 1:
                        if asoc_parent == ASSOCIATIVITY.L_TO_R:
                            # post fix
                            left = parent.ops[0]
                        else:
                            assert asoc_parent == ASSOCIATIVITY.R_TO_L, asoc_parent
                            right = parent.ops[0]
                    else:
                        if i == 0:
                            right = parent.ops[1]
                        else:
                            left = parent.ops[i - 1]
                            if argc > i + 2:
                                right = parent.ops[i + 1]

                    if self._visit_operand_parentheses_extra_check(
                            op_my, precedence_my, asoc_my, parent.fn,
                            precedence_parent, asoc_parent, left, right):
                        use_parenthesis = True
                    else:
                        if left is not None:  # "operand" is on right side of parent operator
                            # same precedence -> parenthesis on right (this) if it is expression
                            # a + (b + c)
                            # a + b + c = (a + b) + c
                            # right with lower precedence -> parenthesis for right not required
                            # a + b * c = a + (b * c)
                            # right with higher precedence -> parenthesis for right
                            # a * (b + c)
                            if precedence_my > precedence_parent:
                                use_parenthesis = True
                            elif precedence_my == precedence_parent:
                                use_parenthesis = argc != 1 or asoc_parent != ASSOCIATIVITY.L_TO_R

                        if not use_parenthesis and right is not None:
                            # "operand" is on left side of parent operator
                            # if op_my == parent.fn:
                            #    right_prec, _, right_op = self._precedence_of_expr(right)
                            #    if right_op == op_my:
                            #        # right and left with same precedence -> parenthesis on both sides
                            #        # (a + b) + (c + d)
                            #        use_parenthesis = True
                            if precedence_my > precedence_parent:
                                # left with higher precedence -> parenthesis for left
                                # (a + b) * c
                                # a + b + c + d = (a + b) + c + d
                                # = ((a + b) + c) + d
                                use_parenthesis = True

        w = self.out.write
        if use_parenthesis:
            w("(")
        self.visit_iHdlExpr(operand)
        if use_parenthesis:
            w(")")

    def _visit_bin_op(self, operator, op_str,
                      expr_requires_parenthesis=False,
                      cancel_parenthesis=False):
        """
        :type operator: HdlOp
        :type op_str: str
        """
        op0, op1 = operator.ops
        self._visit_operand(op0, 0, operator, expr_requires_parenthesis,
                            cancel_parenthesis)
        self.out.write(op_str)
        self._visit_operand(op1, 1, operator, expr_requires_parenthesis,
                            cancel_parenthesis)

    def _visit_operator_index(self, operator):
        """
        :type operator: HdlOp
        """
        op0, op1 = operator.ops
        self._visit_operand(op0, 0, operator, False, False)
        w = self.out.write
        w("[")
        self._visit_operand(op1, 1, operator, False, True)
        w("]")

    def visit_operator_call(self, o):
        """
        :type operator: HdlOp
        """
        self._visit_operand(o.ops[0], 0, o, False, False)
        w = self.out.write
        w("(")
        for is_last, (o_i, _o) in iter_with_last(enumerate(o.ops[1:])):
            self._visit_operand(_o, o_i, o, False, True)
            if not is_last:
                w(", ")
        w(")")

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        raise TypeError("does not support HdlFunctionDef", self, o)

    def visit_HdlStmProcess(self, o):
        """
        :type proc: HdlStmProcess
        """
        raise TypeError("does not support HdlStmProcess", self, o)

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        raise TypeError("does not support HdlStmBlock", self, o)

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        raise TypeError("does not support HdlStmIf", self, o)

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        raise TypeError("does not support HdlStmCase", self, o)

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        raise TypeError("does not support HdlStmWait", self, o)

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        raise TypeError("does not support HdlStmFor", self, o)

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        """
        raise TypeError("does not support HdlStmForIn", self, o)

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        """
        raise TypeError("does not support HdlStmWhile", self, o)

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        raise TypeError("does not support HdlStmAssign", self, o)

    def visit_HdlStmRepeat(self, o):
        """
        :type o: HdlStmRepeat
        """
        raise TypeError("does not support HdlStmRepeat", self, o)

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        raise TypeError("does not support HdlStmReturn", self, o)

    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        raise TypeError("does not support HdlStmContinue", self, o)

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        raise TypeError("does not support HdlStmBreak", self, o)




from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.wrap_module_statements_to_processes import collect_hdl_ids
from hdlConvertorAst.hdlAst import HdlIdDef

def all_ids_constant(expr):
    all_ids = set()
    collect_hdl_ids(expr, all_ids)
    for hid in all_ids:
        if isinstance(hid.obj, HdlIdDef) and not hid.obj.is_const:
            return False
    return True

class DetectCompileTimeStatements(HdlAstVisitor):
    """
    Mark all statements which can be resolved compiletime with a in_preproc flag

    :attentions: requires all symbols to be resolved
    """
    def visit_param(self, o):
        """
        :type o: HdlIdDef
        """
        o.is_const = True
        return o

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        if not o.in_preproc:
            o.in_preproc = all_ids_constant(o.switch_on)

        return HdlAstVisitor.visit_HdlStmCase(self, o)

    def visit_iHdlExpr(self, o):
        """
        :note: do not descend in to expressions because we do not need it
        """
        return o

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        if not o.in_preproc:
            in_preproc = True
            in_preproc &= all_ids_constant(o.cond)
            if in_preproc:
                for c, _ in o.elifs:
                    in_preproc &= all_ids_constant(c)
                    if not in_preproc:
                        break
            o.in_preproc = in_preproc

        return HdlAstVisitor.visit_HdlStmIf(self, o)


from hdlConvertorAst.translate.common.discover_declarations import DiscoverDeclarations
from hdlConvertorAst.translate.common.name_scope import NameScope
from hdlConvertorAst.translate.common.resolve_names import ResolveNames
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.\
    add_unique_labels_to_all_processes import AddUniqueLabelsToAllProcesses
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.apply_io_scope_to_signal_names import ApplyIoScopeToSignalNames
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .assignment_to_update_assignment import AssignmentToUpdateAssignment
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .discover_stm_outputs import discover_stm_outputs_context
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .verilog_operands_to_basic_hdl_sim_model import BasicHdlSimModelTranslateVerilogOperands
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .verilog_types_to_basic_hdl_sim_model import VerilogTypesToBasicHdlSimModel
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .wrap_module_statements_to_processes import wrap_module_statements_to_processes
from hdlConvertorAst.translate.vhdl_to_verilog import link_module_dec_def


def verilog_to_basic_hdl_sim_model(context):
    """
    :type context: HdlContext
    """
    link_module_dec_def(context)
    name_scope = NameScope.make_top(False)

    DiscoverDeclarations(name_scope).visit_HdlContext(context)
    ResolveNames(name_scope).visit_HdlContext(context)
    wrap_module_statements_to_processes(context)
    BasicHdlSimModelTranslateVerilogOperands().visit_HdlContext(context)
    VerilogTypesToBasicHdlSimModel().visit_HdlContext(context)
    stm_outputs = discover_stm_outputs_context(context)

    AddUniqueLabelsToAllProcesses(name_scope, stm_outputs).context(context)
    AssignmentToUpdateAssignment().visit_HdlContext(context)
    ApplyIoScopeToSignalNames().visit_HdlContext(context)

    return context, stm_outputs, name_scope



from typing import List, Union, Optional

from hdlConvertorAst.hdlAst._bases import iHdlObj, iHdlObjInModule, \
                                       iHdlObjWithName, iHdlStatement
from hdlConvertorAst.hdlAst._expr import iHdlExpr, HdlDirection


class HdlIdDef(iHdlObjWithName, iHdlObjInModule):
    """
    Definition of port/param/type/varialbe etc in HDL

    :ivar ~.name: name of newly defined object
    :ivar ~.type: type of the defined variable (or type etc.)
    :ivar ~.value: initialisation of variable (typedef etc.)
    :ivar ~.is_const: flag if true the value is constants
    :ivar ~.is_latched: flag if true the object corresponds to VHDL variable/Verilog reg
    :ivar ~.is_static: flag if true static (same meaning as in C/C++)
    :ivar ~.is_shared: flag if true the variable is VHDL shared variable
    :ivar ~.is_virtual: flag if true the variable stores virtual type
        (corresponds to System Verilog virtual parameter)
    :ivar ~.direction: direction if the variable is port
    """
    __slots__ = ["name", "type", "value", "is_latched", "is_const", "is_static",
                 "is_shared", "is_virtual", "direction"]

    def __init__(self):
        iHdlObjWithName.__init__(self)
        iHdlObjInModule.__init__(self)
        self.type = None  # type: HdlType
        self.value = None  # iHdlExpr
        self.is_const = False  # type: bool
        self.is_latched = False  # type: bool
        self.is_static = False  # type: bool
        self.is_virtual = False  # type: bool
        self.is_shared = False  # type: bool
        self.direction = None  # type: HdlDirection


class HdlFunctionDef(iHdlObjWithName, iHdlObjInModule):
    """
    HDL Function definition or declaration

    :note: Verilog:
        * task automatic corresponds to is_static=False, is_task=True
        * task corresponds to is_static=True, is_task=True
        * task/function in module is static by default,
          if in class it is automatic by default
    """
    __slots__ = [
        "is_declaration_only",
        "is_operator",
        "is_static",
        "is_task",
        "is_virtual",
        "return_t", "params", "body"]

    def __init__(self):
        iHdlObjWithName.__init__(self)
        iHdlObjInModule.__init__(self)
        self.is_declaration_only = True  # type:bool
        self.is_operator = False  # type: bool
        self.is_static = False  # type: bool
        self.is_task = False  # type: bool
        self.is_virtual = False  # type: bool
        self.return_t = None  # type: Optional[iHdlExpr]
        self.params = []  # type: List[HdlIdDef]
        self.body = []  # type: List[Union[HdlIdDef, iHdlStatement, iHdlExpr]]




from typing import Union

from hdlConvertorAst.hdlAst import HdlOp, HdlOpType, iHdlTypeDef, iHdlExpr, \
    HdlTypeBitsDef, HdlValueId, HdlIdDef
from hdlConvertorAst.to.basic_hdl_sim_model.utils import _verilog_slice_to_width
from hdlConvertorAst.to.hwt.utils import BitsT
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.verilog_types_to_basic_hdl_sim_model import VerilogTypesToBasicHdlSimModel


class VerilogTypesToHwt(VerilogTypesToBasicHdlSimModel):

    def visit_HdlModuleDef(self, o):
        """
        Remove genvar instances as they do not need forward declarations

        :type o: HdlModuleDef
        """
        genvar = HdlValueId("genvar")
        o.objs = [_o for _o in o.objs if not isinstance(_o, HdlIdDef) or _o.type != genvar ]
        super(VerilogTypesToHwt, self).visit_HdlModuleDef(o)
        return o

    def _visit_type(self, t):
        """
        :type t: Union[iHdlExpr, iHdlTypeDef]
        """
        if isinstance(t, HdlOp) and t.fn == HdlOpType.INDEX:
            o0, o1 = t.ops
            return hdl_index(self._visit_type(o0), o1)
        elif isinstance(t, HdlTypeBitsDef):
            width = _verilog_slice_to_width(t.msb, t.lsb)
            if width == 1:
                return HdlValueId("BIT")
            return BitsT(width, t.signed)
        elif isinstance(t, HdlValueId):
            if t == HdlValueId("integer"):
                return HdlValueId("INT")
            elif t == HdlValueId("real"):
                return HdlValueId("FLOAT64")
            elif t == HdlValueId("time"):
                return HdlValueId("int")
        raise NotImplementedError(t)



from hdlConvertorAst.hdlAst._expr import HdlOpType, HdlValueId, HdlValueInt, \
    HdlAll, HdlOp, HdlOthers
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.common import ToHdlCommon, ASSOCIATIVITY
from hdlConvertorAst.to.hdlUtils import iter_with_last, Indent
from copy import copy

L = ASSOCIATIVITY.L_TO_R
R = ASSOCIATIVITY.R_TO_L


# https://www.csee.umbc.edu/portal/help/VHDL/operator.html
class ToVhdl2008Expr(ToHdlCommon):
    GENERIC_UNARY_OPS = copy(ToHdlCommon.GENERIC_UNARY_OPS)
    GENERIC_UNARY_OPS.update({
        HdlOpType.NEG: "NOT ",
        HdlOpType.NEG_LOG: "NOT ",
        HdlOpType.RANGE: "RANGE ",  # used in HdlPhysicalDef
    })
    BITWISE_BIN_OPS = {
        HdlOpType.AND, HdlOpType.AND_LOG,
        HdlOpType.OR, HdlOpType.OR_LOG,
        HdlOpType.XOR, HdlOpType.NAND,
        HdlOpType.NOR, HdlOpType.XNOR
    }
    GENERIC_BIN_OPS = {
        HdlOpType.AND: " AND ",
        HdlOpType.AND_LOG: " AND ",
        HdlOpType.OR: " OR ",
        HdlOpType.OR_LOG: " OR ",
        HdlOpType.DIV: " / ",
        HdlOpType.MOD: " MOD ",
        HdlOpType.REM: " REM ",
        HdlOpType.POW: " ** ",
        HdlOpType.NAND: " NAND ",
        HdlOpType.NOR: " NOR ",
        HdlOpType.XOR: " XOR ",
        HdlOpType.XNOR: " XNOR ",
        HdlOpType.EQ: ' = ',
        HdlOpType.NE: " /= ",
        HdlOpType.EQ_MATCH: ' ?= ',
        HdlOpType.NE_MATCH: ' ?/= ',
        HdlOpType.LT_MATCH: ' ?< ',
        HdlOpType.LE_MATCH: ' ?<= ',
        HdlOpType.GT_MATCH: ' ?> ',
        HdlOpType.GE_MATCH: ' ?>= ',
        HdlOpType.TO: " TO ",
        HdlOpType.DOWNTO: " DOWNTO ",
        HdlOpType.ARROW: " => ",
        HdlOpType.MAP_ASSOCIATION: " => ",
        HdlOpType.RANGE: " RANGE ",
        HdlOpType.CONCAT: " & ",
        HdlOpType.ROL: " ROL ",
        HdlOpType.ROR: " ROR ",
        HdlOpType.SLA: " SLA ",
        HdlOpType.SRA: " SRA ",
        HdlOpType.SLL: " SLL ",
        HdlOpType.SRL: " SRL ",
        HdlOpType.UNIT_SPEC: " ",
    }
    EQ_NEQ_OPS = (
        HdlOpType.EQ,
        HdlOpType.NE
    )
    GENERIC_BIN_OPS.update(ToHdlCommon.GENERIC_BIN_OPS)
    NUM_BASES = {
        2: "",
        8: "O",
        16: "X",
        256: "",  # 'X' literals
    }
    OP_PRECEDENCE = {
        HdlOpType.UNIT_SPEC: (0, L),
        HdlOpType.DOT: (1, L),
        HdlOpType.CALL: (1, L),
        HdlOpType.INDEX: (1, L),
        HdlOpType.RISING: (1, L),
        HdlOpType.FALLING: (1, L),
        HdlOpType.APOSTROPHE: (1, L),

        HdlOpType.POW: (2, R),
        HdlOpType.ABS: (2, L),
        HdlOpType.NEG: (2, R),
        HdlOpType.NEG_LOG: (2, R),

        HdlOpType.MUL: (3, L),
        HdlOpType.DIV: (3, L),
        HdlOpType.MOD: (3, L),
        HdlOpType.REM: (3, L),

        HdlOpType.PLUS_UNARY: (4, R),
        HdlOpType.MINUS_UNARY: (4, R),

        HdlOpType.CONCAT: (5, L),
        HdlOpType.ADD: (5, L),
        HdlOpType.SUB: (5, L),

        HdlOpType.SLL: (6, L),
        HdlOpType.SRL: (6, L),
        HdlOpType.SLA: (6, L),
        HdlOpType.SRA: (6, L),
        HdlOpType.ROL: (6, L),
        HdlOpType.ROR: (6, L),

        HdlOpType.EQ: (7, L),
        HdlOpType.NE: (7, L),
        HdlOpType.GT: (7, L),
        HdlOpType.LT: (7, L),
        HdlOpType.GE: (7, L),
        HdlOpType.LE: (7, L),
        HdlOpType.EQ_MATCH: (7, L),
        HdlOpType.NE_MATCH: (7, L),
        HdlOpType.LT_MATCH: (7, L),
        HdlOpType.LE_MATCH: (7, L),
        HdlOpType.GT_MATCH: (7, L),
        HdlOpType.GE_MATCH: (7, L),

        HdlOpType.AND: (8, L),
        HdlOpType.OR: (8, L),
        HdlOpType.NAND: (8, ASSOCIATIVITY.NONE),
        HdlOpType.NOR: (8, ASSOCIATIVITY.NONE),
        HdlOpType.XOR: (8, L),
        HdlOpType.XNOR: (8, ASSOCIATIVITY.NONE),

        HdlOpType.DOWNTO: (9, L),
        HdlOpType.TO: (9, L),
        HdlOpType.TERNARY: (9, R),

        HdlOpType.RANGE: (10, L),
        HdlOpType.ARROW: (11, L),
        HdlOpType.MAP_ASSOCIATION: (11, L),
    }

    def visit_HdlValueInt(self, o):
        w = self.out.write
        v = o.val
        bits = o.bits
        if is_str(v):
            v = v.upper()
            if o.base == 256:
                w("'%s'" % v)
            else:
                w('%s"%s"' % (self.NUM_BASES[o.base], v))
            return

        if bits is None:
            if o.base is not None:
                b = self.NUM_BASES[o.base]
                if o.base == 256:
                    w("'%d'" % v)
                elif o.base == 16:
                    w('%s"%X"' % (b, v))
                elif o.base == 8:
                    w('%s"%o"' % (b, v))
                elif o.base == 2:
                    w('{0}"{1:b}"'.format(b, v))
                else:
                    raise NotImplementedError(o.base)
            else:
                w(str(v))
            return
        elif bits % 8 == 0:
            f = 'X"{0:0%dx}"' % (bits / 8)
        else:
            f = '"{0:0%db}"' % (bits)
        w(f.format(v))

    def _visit_operand_parentheses_extra_check(self, op_my, precedence_my, asoc_my,
                                               op_parent, precedence_parent, asoc_parent,
                                               left, right):
        if super(ToVhdl2008Expr, self)._visit_operand_parentheses_extra_check(
                op_my, precedence_my, asoc_my,
                op_parent, precedence_parent, asoc_parent, left, right):
            return True
        elif op_my in self.BITWISE_BIN_OPS and\
                op_parent in self.BITWISE_BIN_OPS and\
                op_my != op_parent:
            return True
        else:
            return False

    def _visit_operand(self, operand, i,
        parent,
        expr_requires_parenthesis,
        cancel_parenthesis):
        if parent.fn in self.EQ_NEQ_OPS and isinstance(operand, HdlOp) and operand.fn in self.EQ_NEQ_OPS:
            # handle the case (a = b) = (c = d) where to boolean conversion is applied during a = b evaluation
            expr_requires_parenthesis = True

        return ToHdlCommon._visit_operand(self, operand, i, parent, expr_requires_parenthesis, cancel_parenthesis)

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        """
        fn = o.ops[0]
        if fn == HdlValueId("assert"):
            self.visit_assert(o.ops[1:])
            return
        elif fn == HdlValueId("report"):
            self.visit_report(o.ops[1:])
            return

        w = self.out.write
        op = o.fn
        if op == HdlOpType.RISING:
            w("RISING_EDGE(")
            self.visit_iHdlExpr(o.ops[0])
            w(")")
        elif op == HdlOpType.FALLING:
            w("FALLING_EDGE(")
            self.visit_iHdlExpr(o.ops[0])
            w(")")
        elif op == HdlOpType.INDEX or op == HdlOpType.CALL:
            self._visit_operand(o.ops[0], 0, o, False, False)
            w("(")
            for isLast, (o_i, _o) in iter_with_last(enumerate(o.ops[1:])):
                self._visit_operand(_o, o_i + 1, o, False, True)
                if not isLast:
                    w(", ")
            w(")")
        elif op == HdlOpType.TERNARY:
            has_3_ops = len(o.ops) == 3
            if has_3_ops:
                cond, o0, o1 = o.ops
            else:
                cond, o0 = o.ops

            self._visit_operand(o0, 1, o, True, False)
            w(" WHEN ")
            self._visit_operand(cond, 0, o, True, False)
            if has_3_ops:
                if isinstance(o1, HdlOp) and o1.fn == HdlOpType.TERNARY:
                    w(" ELSE\n")
                    self.visit_iHdlExpr(o1)  # must not have parenthesis
                else:
                    w(" ELSE ")
                    self._visit_operand(o1, 2, o, False, False)
        elif op == HdlOpType.APOSTROPHE:
            self._visit_operand(o.ops[0], 0, o, True, False)
            w("'")
            args = o.ops[1]
            if isinstance(args, list):
                self.visit_iHdlExpr(args)
            elif isinstance(args, HdlValueId):
                # normal attribute
                self.visit_iHdlExpr(args)
            else:
                w("(")
                self._visit_operand(args, 0, o, False, True)
                w(")")
        elif op == HdlOpType.ABS:
            w("ABS(")
            self.visit_iHdlExpr(o.ops[0])
            w(")")
        elif op == HdlOpType.DEFINE_RESOLVER:
            assert self.in_typedef
            self.visit_iHdlExpr(o.ops[0])
            w(" ")
            self.visit_iHdlExpr(o.ops[1])
        else:
            return ToHdlCommon.visit_HdlOp(self, o)

    def visit_str(self, o):
        """
        :type o: str
        """
        w = self.out.write
        if o == "":
            w('""')
        else:
            ESCAPES = {
                '\n': 'LF\n',
                '\c': 'CR'
            }
            CONC = self.GENERIC_BIN_OPS[HdlOpType.CONCAT]
            first = True
            in_string_lit = False
            for c in o:
                esc = ESCAPES.get(c, None)
                if esc is None:
                    if not in_string_lit:
                        if not first:
                            # first character after escape sequence
                            w(CONC)
                        w('"')  # string start "
                        in_string_lit = True
                    # character inside ""
                    w(c)
                else:
                    if not first:
                        if in_string_lit:
                            w('"')  # string end "
                            # escape sequence behind string
                        w(CONC)
                    in_string_lit = False
                    # escape name
                    w(esc)
                first = False

            if in_string_lit:
                w('"')  # string end "

    def visit_iHdlExpr(self, expr):
        w = self.out.write
        if expr is HdlAll:
            w("ALL")
        elif expr is HdlOthers:
            w("OTHERS")
        elif expr is None:
            if self.in_typedef:
                w("<>")
            else:
                w("OPEN")
        elif is_str(expr):
            self.visit_str(expr)
        elif isinstance(expr, list):
            with_nl = len(expr) > 3
            if with_nl:
                w("(\n")
            else:
                w("(")
            with Indent(self.out):
                for is_last, elem in iter_with_last(expr):
                    self.visit_iHdlExpr(elem)
                    if not is_last:
                        if with_nl:
                            w(",\n")
                        else:
                            w(", ")
            w(")")
        else:
            ToHdlCommon.visit_iHdlExpr(self, expr)



from typing import Set, List, Dict

from hdlConvertorAst.hdlAst import iHdlExpr, HdlValueId, \
    HdlContext, HdlModuleDec, HdlStmAssign, HdlStmProcess, \
    HdlStmIf, HdlStmBlock
from hdlConvertorAst.hdlAst._structural import HdlModuleDef
from hdlConvertorAst.hdlAst._expr import HdlOp, HdlOpType


def get_output_ids(e, outputs):
    """
    :type e: iHdlExpr
    :type outputs: Set[HdlValueId]
    """
    if isinstance(e, tuple):
        for o in e:
            get_output_ids(o, outputs)
        return  
    cur_path = []
    while isinstance(e, HdlOp):
        if not e.fn == HdlOpType.DOT:
            raise NotImplementedError(e)
        op0, op1 = e.ops
        assert isinstance(op1, HdlValueId), op1
        cur_path.append(op1)
        e = op0
    if isinstance(e, HdlValueId):
        cur_path.append(e)
        if len(cur_path) == 1:
            cur_path = cur_path[0]
        else:
            cur_path = tuple(reversed(cur_path))

        outputs.add(cur_path)
    else:
        raise NotImplementedError(e)


def discover_outputs(stm, outputs):
    """
    :type stm: iHdlStm
    :type outputs: Dict["HdlStm", Set[HdlValueId]]
    """
    _outputs = None
    if isinstance(stm, HdlStmAssign):
        o = set()
        get_output_ids(stm.dst, o)
    elif isinstance(stm, HdlStmBlock):
        _outputs = {}
        for obj in stm.body:
            discover_outputs(obj, _outputs)
    elif isinstance(stm, HdlStmIf):
        _outputs = {}
        if stm.if_true:
            discover_outputs(stm.if_true, _outputs)
        for (_, elif_case) in stm.elifs:
            if elif_case:
                discover_outputs(elif_case, _outputs)
        if stm.if_false:
            discover_outputs(stm.if_false, _outputs)
    elif isinstance(stm, HdlStmProcess):
        o = discover_outputs(stm.body, outputs)
    else:
        raise NotImplementedError(stm)

    if _outputs is not None:
        # collect outputs from substatements and update main outputs
        o = set()
        for os in _outputs.values():
            for _o in os:
                assert isinstance(_o, (HdlValueId, tuple)), _o
                o.add(_o)
        outputs.update(_outputs)

    outputs[stm] = o
    return o


def discover_stm_outputs(stm):
    """
    :type stm: iHdlStm
    :return: Dict["HdlStm", List[HdlValueId]]
    """
    _outputs = {}
    # :type _outputs: Dict[HdlStm, Set[HdlValueId]]
    discover_outputs(stm, _outputs)
    _outputs = {
        k: list(sorted(outputs, key=lambda x: (isinstance(x, HdlValueId), x)))
        for k, outputs in _outputs.items()
    }
    return _outputs


def discover_stm_outputs_context(c):
    """
    :type c: HdlContext
    """
    outputs = {}
    # :type _outputs: Dict[HdlStm, Set[HdlValueId]]
    for o0 in c.objs:
        if isinstance(o0, HdlModuleDef):
            for o1 in o0.objs:
                if isinstance(o1, HdlStmProcess):
                    _outputs = discover_stm_outputs(o1)
                    outputs.update(_outputs)
    return outputs



from hdlConvertorAst.to.json import ToJson
from hdlConvertorAst.hdlAst._bases import iHdlObj
from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlDirection


class ToJsonDebug(ToJson):
    """
    HdlConverto AST -> json (dict/list/str/int/None composed object)
    An invalid object are converted to str using its __repr__()
    """
    def visit_HdlIdDef(self, o):
        try:
            return ToJson.visit_HdlIdDef(self, o)
        except Exception:
            if isinstance(o, HdlIdDef):
                raise
            return repr(o)

    def visit_HdlDirection(self, o):
        if not isinstance(o, HdlDirection):
            return repr(o)
        else:
            return super(ToJsonDebug, self).visit_HdlDirection(o)

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        try:
            return super(ToJsonDebug, self).visit_iHdlExpr(o)
        except Exception:
            if o.__class__.__repr__ is iHdlObj.__repr__:
                # in order to prevent infinite loop if there is something
                # wrong in serializer code itself
                raise
            return repr(o)



from hdlConvertorAst.hdlAst._expr import HdlValueInt
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.to.verilog.utils import collect_array_dims
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr


class AssignmentToUpdateAssignment(HdlAstVisitor):
    """
    Convert an assignment to an assignment of a update.
    Assignment of a update has folowing format:

    .. code-block:: python

        # dst is a name of destination signal
        # src is name of source signal or value
        # is_clock_edge_dependent is a flag which tells if
        # the assignment is updating some clock edge memory

        # scalar updater
        dst.val_next = (src, is_clock_edge_dependent)

        # updater for array element
        dst.val_next = (src, indexes, is_clock_edge_dependent)

    """

    def pop_index_list(self, o):
        """
        :type o: iHdlExpr
        """
        return collect_array_dims(o)

    def is_clock_edge_dependent(self, o):
        """
        :type o: HdlStmAssign
        """
        # [TODO]
        return HdlValueInt(0, 1, None)

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        assert o.event_delay is None, o
        assert o.time_delay is None, o
        is_clock_edge_dependent = self.is_clock_edge_dependent(o)
        dst, dst_index = self.pop_index_list(o.dst)
        o.dst = hdl_getattr(dst, "val_next")
        if dst_index:
            o.src = [o.src, dst_index, is_clock_edge_dependent]
        else:
            o.src = [o.src,  is_clock_edge_dependent]



from hdlConvertorAst.hdlAst import HdlOp, HdlStmIf, HdlStmBlock, HdlStmAssign
from hdlConvertorAst.to.basic_hdl_sim_model.expr import ToBasicHdlSimModelExpr
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last


class ToBasicHdlSimModelStm(ToBasicHdlSimModelExpr):

    def visit_HdlStmProcess(self, proc):
        """
        :type proc: HdlStmProcess
        """
        w = self.out.write
        w("# sensitivity: ")
        for last, s in iter_with_last(proc.sensitivity):
            if isinstance(s, HdlOp):
                w(str(s.fn))
                w(" ")
                self.visit_iHdlExpr(s.ops[0])
            else:
                self.visit_iHdlExpr(s)
            if not last:
                w(", ")
        w("\n")
        w("def ")
        w(proc.labels[0])
        w("(self):\n")
        if proc.trigger_constrain is not None:
            raise NotImplementedError()
        body = proc.body
        with Indent(self.out):
            self.visit_iHdlStatement_in_statement(body)
        w("\n")

    def visit_iHdlStatement_in_statement(self, stm):
        if isinstance(stm, HdlStmAssign):
            self.visit_HdlStmAssign(stm)
        elif isinstance(stm, HdlStmIf):
            self.visit_HdlStmIf(stm)
        elif isinstance(stm, HdlStmBlock):
            self.visit_HdlStmBlock(stm)
        else:
            raise NotImplementedError(stm.__class__, stm)

    def visit_HdlStmBlock(self, stm):
        """
        :type stm: HdlStmBlock
        """
        w = self.out.write
        if not stm.body:
            self.out.write("pass")
            return

        for is_last, i in iter_with_last(stm.body):
            self.visit_iHdlStatement_in_statement(i)
            if not is_last:
                w("\n")

    def visit_HdlStmIf(self, stm):
        """
        :type stm: HdlStmIf

        if cond:
            ...
        else:
            ...
        """
        w = self.out.write
        c = stm.cond
        ifTrue = stm.if_true
        ifFalse = stm.if_false
        w("if ")
        self.visit_iHdlExpr(c)
        w(":\n")
        with Indent(self.out):
            self.visit_iHdlStatement_in_statement(ifTrue)
            w("\n")

        for (c, _stm) in stm.elifs:
            w("elif ")
            self.visit_iHdlExpr(c)
            w(":\n")
            with Indent(self.out):
                self.visit_iHdlStatement_in_statement(_stm)
                w("\n")

        w("else:\n")
        with Indent(self.out):
            if ifFalse is None:
                w("pass")
            else:
                self.visit_iHdlStatement_in_statement(ifFalse)

    def visit_HdlStmAssign(self, a):
        """
        :type a: HdlStmAssign
        """
        w = self.out.write
        self.visit_iHdlExpr(a.dst)
        w(" = ")
        # if a.is_blocking:
        #     raise NotImplementedError(a)
        if a.time_delay is not None:
            raise NotImplementedError()
        if a.event_delay is not None:
            raise NotImplementedError()
        self.visit_iHdlExpr(a.src)

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        self.visit_doc(o)
        w = self.out.write
        w("raise")
        if o.val is not None:
            w(" ")
            self.visit_iHdlExpr(o.val)
        w("\n")

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        self.visit_doc(o)
        w = self.out.write
        w("yield Timer(")
        self.visit_iHdlExpr(o.val)
        w(")")

    def visit_HdlStmNop(self, o):
        """
        :type o: HdlStmNop
        """
        self.visit_doc(o)
        self.out.write("pass")



"""
(System)Verilog language keyword lists
"""

IEEE1364_1995_KEYWORDS = [
    "always",
    "and",
    "assign",
    "begin",
    "buf",
    "bufif0",
    "bufif1",
    "case",
    "casex",
    "casez",
    "cmos",
    "deassign",
    "default",
    "defparam",
    "disable",
    "edge",
    "else",
    "end",
    "endcase",
    "endfunction",
    "endmodule",
    "endprimitive",
    "endspecify",
    "endtable",
    "endtask",
    "event",
    "for",
    "force",
    "forever",
    "fork",
    "function",
    "highz0",
    "highz1",
    "if",
    "ifnone",
    "rpmos",
    "initial",
    "rtran",
    "inout",
    "rtranif0",
    "input",
    "rtranif1",
    "integer",
    "scalared",
    "join",
    "small",
    "large",
    "specify",
    "macromodule",
    "specparam",
    "medium",
    "strong0",
    "module",
    "strong1",
    "nand",
    "supply0",
    "negedge",
    "supply1",
    "nmos",
    "table",
    "nor",
    "task",
    "not",
    "time",
    "notif0",
    "tran",
    "notif1",
    "tranif0",
    "or",
    "tranif1",
    "output",
    "tri",
    "parameter",
    "tri0",
    "pmos",
    "tri1",
    "posedge",
    "triand",
    "primitive",
    "trior",
    "pull0",
    "trireg",
    "pull1",
    "vectored",
    "pulldown",
    "wait",
    "pullup",
    "wand",
    "rcmos",
    "weak0",
    "real",
    "weak1",
    "realtime",
    "while",
    "reg",
    "wire",
    "release",
    "wor",
    "repeat",
    "xnor",
    "rnmos",
    "xor",
]

IEEE1364_2001_KEYWORDS = IEEE1364_1995_KEYWORDS + [
    "automatic",
    "cell",
    "config",
    "incdir",
    "include",
    "instance",
    "liblist",
    "library",
    "localparam",
    "noshowcancelled",
    "pulsestyle_ondetect",
    "design",
    "endconfig",
    "endgenerate",
    "generate",
    "genvar",
    "pulsestyle_onevent",
    "showcancelled",
    "signed",
    "unsigned",
    "use",
]

IEEE1364_2001_NOCONFIG_KEYWORDS = [
    kw for kw in IEEE1364_2001_KEYWORDS if kw not in {
        "cell",
        "config",
        "design",
        "endconfig",
        "incdir",
        "include",
        "instance",
        "liblist",
        "library",
        "use",
    }
]

IEEE1364_2005_KEYWORDS = IEEE1364_2001_KEYWORDS + ["uwire"]

IEEE1800_2005_KEYWORDS = IEEE1364_2005_KEYWORDS + [
    "alias",
    "endsequence",
    "pure",
    "always_comb",
    "enum",
    "rand",
    "always_ff",
    "expect",
    "randc",
    "always_latch",
    "export",
    "randcase",
    "assert",
    "extends",
    "randsequence",
    "assume",
    "extern",
    "ref",
    "before",
    "final",
    "return",
    "bind",
    "first_match",
    "sequence",
    "bins",
    "foreach",
    "shortint",
    "binsof",
    "forkjoin",
    "shortreal",
    "bit",
    "iff",
    "solve",
    "break",
    "ignore_bins",
    "static",
    "byte",
    "illegal_bins",
    "string",
    "chandle",
    "import",
    "struct",
    "class",
    "inside",
    "super",
    "clocking",
    "int",
    "tagged",
    "const",
    "interface",
    "this",
    "constraint",
    "intersect",
    "throughout",
    "context",
    "join_any",
    "timeprecision",
    "continue",
    "join_none",
    "timeunit",
    "cover",
    "local",
    "type",
    "covergroup",
    "logic",
    "typedef",
    "coverpoint",
    "longint",
    "union",
    "cross",
    "matches",
    "unique",
    "dist",
    "modport",
    "var",
    "do",
    "new",
    "virtual",
    "endclass",
    "null",
    "void",
    "endclocking",
    "package",
    "wait_order",
    "endgroup",
    "packed",
    "wildcard",
    "endinterface",
    "priority",
    "with",
    "endpackage",
    "program",
    "within",
    "endprogram",
    "property",
    "endproperty",
    "protected",
]

IEEE1800_2009_KEYWORDS = IEEE1800_2005_KEYWORDS + [
    "accept_on",
    "checker",
    "endchecker",
    "eventually",
    "global",
    "implies",
    "let",
    "nexttime",
    "reject_on",
    "restrict",
    "s_always",
    "s_eventually",
    "s_nexttime",
    "s_until",
    "s_until_with",
    "strong",
    "sync_accept_on",
    "sync_reject_on",
    "unique0",
    "until",
    "until_with",
    "untyped",
    "weak",
]

IEEE1800_2012_KEYWORDS = IEEE1800_2009_KEYWORDS + [
    "implements",
    "nettype",
    "interconnect",
    "soft",
]

IEEE1800_2017_KEYWORDS = IEEE1800_2012_KEYWORDS + [
]

IEEE1800_2017_DOLAR_SYMBOLS = [
    #Simulation control tasks (20.2) 
    '$finish',
    '$stop',
    '$exit',
    # Simulation time functions (20.3)
    '$realtime',
    '$stime',
    '$time',
    # Timescale tasks (20.4)',
    '$printtimescale',
    '$timeformat',
    # Conversion functions (20.5)',
    '$bitstoreal',
    '$realtobits',
    '$bitstoshortreal',
    '$shortrealtobits',
    '$itor',
    '$rtoi',
    '$signed',
    '$unsigned',
    '$cast',
    # Data query functions (20.6)
    '$bits',
    '$isunbounded',
    '$typename',
    # Array query functions (20.7)
    '$unpacked_dimensions',
    '$dimensions',
    '$left',
    '$right',
    '$low',
    '$high',
    '$increment'
    '$size',
    # Math functions (20.8)
    '$clog2',
    '$asin',
    '$ln',
    '$acos',
    '$log10',
    '$atan',
    '$exp',
    '$atan2',
    '$sqrt',
    '$hypot',
    '$pow',
    '$sinh',
    '$floor',
    '$cosh',
    '$ceil',
    '$tanh',
    '$sin',
    '$asinh',
    '$cos',
    '$acosh',
    '$tan',
    '$atanh',
    # Bit vector system functions (20.9)
    '$countbits',
    '$countones',
    '$onehot',
    '$onehot0',
    '$isunknown',
    # Severity tasks (20.10)
    # Elaboration tasks (20.11)
    '$fatal',
    '$error',
    '$warning',
    '$info',
    # Assertion control tasks (20.12)
    '$asserton',
    '$assertoff',
    '$assertkill',
    '$assertcontrol',
    '$assertpasson',
    '$assertpassoff',
    '$assertfailon',
    '$assertfailoff',
    '$assertnonvacuouson',
    '$assertvacuousoff',
    # Sampled value system functions (20.13)
    '$sampled',
    '$rose',
    '$fell',
    '$stable',
    '$changed',
    '$past',
    '$past_gclk',
    '$rose_gclk',
    '$fell_gclk',
    '$stable_gclk',
    '$changed_gclk',
    '$future_gclk',
    '$rising_gclk',
    '$falling_gclk',
    '$steady_gclk',
    '$changing_gclk',
    # Coverage control functions (20.14)
    '$coverage_control',
    '$coverage_get_max',
    '$coverage_get',
    '$coverage_merge',
    '$coverage_save',
    '$get_coverage',
    '$set_coverage_db_name',
    '$load_coverage_db',
    # Probabilistic distribution functions (20.15)
    '$random',
    '$dist_chi_square',
    '$dist_erlang',
    '$dist_exponential',
    '$dist_normal',
    '$dist_poisson',
    '$dist_t',
    '$dist_uniform',
    # Stochastic analysis tasks and functions (20.16)
    '$q_initialize',
    '$q_add',
    '$q_remove',
    '$q_full',
    '$q_exam',
    # PLA modeling tasks (20.17)
    '$async$and$array',
    '$async$and$plane',
    '$async$nand$array',
    '$async$nand$plane',
    '$async$or$array',
    '$async$or$plane',
    '$async$nor$array',
    '$async$nor$plane',
    '$sync$and$array',
    '$sync$and$plane',
    '$sync$nand$array',
    '$sync$nand$plane',
    '$sync$or$array',
    '$sync$or$plane',
    '$sync$nor$array',
    '$sync$nor$plane',
    # Miscellaneous tasks and functions (20.18)
    '$system',
    #Display tasks (21.2)
    '$display',
    '$write',
    '$displayb',
    '$writeb',
    '$displayh',
    '$writeh ',
    '$displayo',
    '$writeo',
    '$strobe',
    '$monitor ',
    '$strobeb',
    '$monitorb',
    '$strobeh',
    '$monitorh',
    '$strobeo',
    '$monitoro',
    '$monitoroff',
    '$monitoron',
    # File I/O tasks and functions (21.3)
    '$fclose',
    '$fopen',
    '$fdisplay',
    '$fwrite',
    '$fdisplayb',
    '$fwriteb',
    '$fdisplayh',
    '$fwriteh ',
    '$fdisplayo',
    '$fwriteo',
    '$fstrobe',
    '$fmonitor ',
    '$fstrobeb',
    '$fmonitorb',
    '$fstrobeh',
    '$fmonitorh',
    '$fstrobeo',
    '$fmonitoro',
    '$swrite',
    '$sformat',
    '$swriteb',
    '$sformatf ',
    '$swriteh',
    '$fgetc',
    '$swriteo',
    '$ungetc',
    '$fscanf',
    '$fgets',
    '$fread',
    '$sscanf',
    '$fseek',
    '$rewind',
    '$fflush',
    '$ftell',
    '$feof ',
    '$ferror',
    # Memory load tasks (21.4)
    '$readmemb',
    '$readmemh',
    # Memory dump tasks (21.5)
    '$writememb',
    '$writememh',
    # Command line input (21.6)
    '$test',
    '$value',
    '$plusargs',
    # VCD tasks (21.7)
    '$dumpfile',
    '$dumpvars',
    '$dumpoff',
    '$dumpon',
    '$dumpall',
    '$dumplimit',
    '$dumpflush',
    '$dumpports',
    '$dumpportsoff',
    '$dumpportson',
    '$dumpportsall',
    '$dumpportslimit',
    '$dumpportsflush',
    
]


from hdlConvertorAst.hdlAst import HdlDirection, HdlOpType, \
    HdlOp, HdlCompInst, HdlIdDef, iHdlStatement, \
    HdlTypeAuto, HdlFunctionDef
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last
from hdlConvertorAst.to.verilog.stm import ToVerilog2005Stm


class ToVerilog2005(ToVerilog2005Stm):
    """
    Convert HdlConverotr hdlObject AST back to Verilog 2002
    """
    DIR2V = {
        HdlDirection.IN: "input",
        HdlDirection.OUT: "output",
        HdlDirection.INOUT: "inout",
    }

    def __init__(self, out_stream):
        ToVerilog2005Stm.__init__(self, out_stream)
        self._type_requires_nettype = True

    def visit_doc(self, obj):
        return super(ToVerilog2005, self).visit_doc(obj, "//")

    def visit_direction(self, d):
        """
        :type d: HdlDirection
        """
        vd = self.DIR2V[d]
        self.out.write(vd)

    def visit_generic_declr(self, g):
        """
        :type g: HdlIdDef
        """
        self.visit_doc(g)
        trnt = self._type_requires_nettype
        try:
            self._type_requires_nettype = False
            w = self.out.write
            w("parameter ")
            if g.type is HdlTypeAuto:
                is_array = False
            else:
                is_array = self.visit_type_first_part(g.type)
                w(" ")
            w(g.name)
            if is_array:
                self.visit_type_array_part(g.type)
            v = g.value
            if v is not None:
                w(" = ")
                self.visit_iHdlExpr(v)
        finally:
            self._type_requires_nettype = trnt

    def visit_port_declr(self, p):
        """
        :type p: HdlIdDef
        """
        w = self.out.write
        self.visit_doc(p)
        self.visit_direction(p.direction)
        w(" ")

        t = p.type
        if t is HdlTypeAuto:
            w("wire ")
            is_array = False
        else:
            is_array = self.visit_type_first_part(t)
            w(" ")

        w(p.name)
        if is_array:
            self.visit_type_array_part(t)

    def visit_HdlIdDef(self, var):
        """
        :type var: HdlIdDef
        """
        self.visit_doc(var)
        name = var.name
        t = var.type
        w = self.out.write

        if var.is_const:
            w("localparam ")
            assert var.value is not None, var.name

        if t is HdlTypeAuto:
            if self._type_requires_nettype and not var.is_const:
                w("wire ")
            is_array = False
        else:
            trnt = self._type_requires_nettype
            try:
                if var.is_const:
                    self._type_requires_nettype = False

                is_array = self.visit_type_first_part(t)
            finally:
                self._type_requires_nettype = trnt
            w(" ")
        w(name)
        if is_array:
            self.visit_type_array_part(t)

        if var.value is not None:
            w(" = ")
            self.visit_iHdlExpr(var.value)

        return True

    def visit_map_item(self, item):
        if isinstance(item, HdlOp)\
                and item.fn == HdlOpType.MAP_ASSOCIATION:
            w = self.out.write
            # k, v pair
            k, v = item.ops
            w(".")
            self.visit_iHdlExpr(k)
            w("(")
            self.visit_iHdlExpr(v)
            w(")")
        else:
            self.visit_iHdlExpr(item)

    def visit_map(self, map_):
        w = self.out.write
        with Indent(self.out):
            for last, m in iter_with_last(map_):
                self.visit_map_item(m)
                if last:
                    w("\n")
                else:
                    w(",\n")

    def visit_HdlCompInst(self, c):
        """
        :type c: HdlCompInst
        """
        self.visit_doc(c)
        w = self.out.write
        assert c.module_name
        self.visit_iHdlExpr(c.module_name)
        gms = c.param_map
        if gms:
            w(" #(\n")
            self.visit_map(gms)
            w(")")
        w(" ")
        w(c.name.val)
        pms = c.port_map
        if pms:
            w(" (\n")
            self.visit_map(pms)
            w(")")
        return True

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        self.visit_doc(o)
        w = self.out.write
        if o.is_task:
            w("task ")
        else:
            w("function ")
        if not o.is_static:
            w("automatic ")

        if not o.is_task:
            trnt = self._type_requires_nettype
            try:
                self._type_requires_nettype = False
                self.visit_type_first_part(o.return_t)
                self.visit_type_array_part(o.return_t)
            finally:
                self._type_requires_nettype = trnt
        if o.is_virtual or o.is_operator:
            raise NotImplementedError(o)
        if not o.is_task and o.return_t is not HdlTypeAuto:
            # because " " is already written by previous string
            w(" ")
        w(o.name)
        ps = o.params
        trnt = self._type_requires_nettype
        try:
            self._type_requires_nettype = False
            if ps:
                w(" (\n")
                with Indent(self.out):
                    for last, p in iter_with_last(ps):
                        self.visit_port_declr(p)
                        if last:
                            w("\n")
                        else:
                            w(",\n")
                w(")")
            w(";\n")
        finally:
            self._type_requires_nettype = trnt
        with Indent(self.out):
            for s in o.body:
                if isinstance(s, HdlIdDef):
                    self.visit_HdlIdDef(s)
                    w(";\n")
                elif isinstance(s, iHdlStatement):
                    need_semi = self.visit_iHdlStatement(s)
                    if need_semi:
                        w(";\n")
                    else:
                        w("\n")
                else:
                    self.visit_iHdlExpr(s)
                    w(";\n")
        if o.is_task:
            w("endtask")
        else:
            w("endfunction")

    def visit_HdlClassDef(self, o):
        raise NotImplementedError()

    def visit_HdlPhysicalDef(self, o):
        raise NotImplementedError()

    def visit_HdlEnumDef(self, o):
        raise NotImplementedError()

    def visit_HdlModuleDec(self, e):
        raise ValueError(self, "does not support a module headers without body")

    def visit_HdlModuleDef(self, a):
        """
        :type a: HdlModuleDef
        """
        mod_dec = a.dec
        assert mod_dec is not None, a
        assert not mod_dec.objs, mod_dec
        self.visit_doc(mod_dec)
        w = self.out.write
        w("module ")
        w(mod_dec.name)
        gs = mod_dec.params
        if gs:
            w(" #(\n")
            with Indent(self.out):
                for last, g in iter_with_last(gs):
                    self.visit_generic_declr(g)
                    if last:
                        w("\n")
                    else:
                        w(",\n")

            w(")")
        ps = mod_dec.ports
        if ps:
            w(" (\n")
            with Indent(self.out):
                for last, p in iter_with_last(ps):
                    self.visit_port_declr(p)
                    if last:
                        w("\n")
                    else:
                        w(",\n")
            w(")")
        w(";\n")

        w = self.out.write
        with Indent(self.out):
            for o in a.objs:
                if isinstance(o, HdlIdDef):
                    self.visit_HdlIdDef(o)
                    w(";\n")
                elif isinstance(o, HdlCompInst):
                    self.visit_HdlCompInst(o)
                    w(";\n\n")
                elif isinstance(o, iHdlStatement):
                    need_semi = self.visit_iHdlStatement(o)
                    if need_semi:
                        w(";\n")
                    else:
                        w("\n\n")
                elif isinstance(o, HdlFunctionDef):
                    self.visit_HdlFunctionDef(o)
                    w("\n\n")
                elif o is None:
                    w(";\n")
                else:
                    raise NotImplementedError(o)

        self.out.write("endmodule\n")




from hdlConvertorAst.hdlAst import HdlOpType, HdlStmWait, HdlStmBlock,\
    HdlStmProcessTriggerConstrain
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last
from hdlConvertorAst.to.verilog.expr import ToVerilog2005Expr


class ToVerilog2005Stm(ToVerilog2005Expr):
    ASSIGN_OPS = {
        HdlOpType.ASSIGN: '=',
        HdlOpType.PLUS_ASSIGN: '+=',
        HdlOpType.MINUS_ASSIGN: '-=',
        HdlOpType.MUL_ASSIGN: '*=',
        HdlOpType.DIV_ASSIGN: '/=',
        HdlOpType.MOD_ASSIGN: '%=',
        HdlOpType.AND_ASSIGN: '&=',
        HdlOpType.OR_ASSIGN: '|=',
        HdlOpType.XOR_ASSIGN: '^=',
        HdlOpType.SHIFT_LEFT_ASSIGN: '<<=',
        HdlOpType.SHIFT_RIGHT_ASSIGN: '>>=',
        HdlOpType.ARITH_SHIFT_LEFT_ASSIGN: '<<<=',
        HdlOpType.ARITH_SHIFT_RIGHT_ASSIGN: '>>>=',
    }

    def __init__(self, out_stream):
        super(ToVerilog2005Stm, self).__init__(out_stream)
        self.top_stm = None
        self.is_in_loop_spec = False

    def visit_iHdlStatement(self, stm):
        """
        :type stm: iHdlStatement
        """
        if self.top_stm is None:
            self.top_stm = stm
            try:
                return ToVerilog2005Expr.visit_iHdlStatement(self, stm)
            finally:
                self.top_stm = None
        else:
            return ToVerilog2005Expr.visit_iHdlStatement(self, stm)

    def visit_HdlStmProcess(self, proc):
        """
        :type proc: HdlStmProcess
        """
        self.visit_doc(proc)
        sens = proc.sensitivity
        body = proc.body
        w = self.out.write
        skip_body = False
        if sens is None and proc.trigger_constrain is None:
            if isinstance(body, HdlStmWait):
                skip_body = True
                wait = body
                body = []
            elif (isinstance(body, HdlStmBlock)
                    and body.body
                    and isinstance(body.body[0], HdlStmWait)):
                wait = body.body[0]
                body = body.body[1:]
            else:
                wait = None

            if wait is None:
                assert self.top_stm is proc
                assert isinstance(body, HdlStmBlock), body
                body = body.body
                wait = body[-1]
                assert isinstance(wait, HdlStmWait), wait
                assert len(wait.val) == 0
                body = body[:-1]
                w("initial")
            else:
                if self.top_stm is proc:
                    w("always ")
                w("#")
                assert len(wait.val) == 1
                self.visit_iHdlExpr(wait.val[0])
            _body = HdlStmBlock()
            _body.body = body
            body = _body
        else:
            tr = proc.trigger_constrain
            if self.top_stm is proc:
                if tr is None:
                    w("always ")
                elif tr is HdlStmProcessTriggerConstrain.FF:
                    w("always_ff ")
                elif tr is HdlStmProcessTriggerConstrain.COMB:
                    w("always_comb ")
                elif tr is HdlStmProcessTriggerConstrain.LATCH:
                    w("always_latch ")
                else:
                    raise ValueError(proc.trigger_constrain)
            if tr is None:
                w("@(")
                for last, item in iter_with_last(sens):
                    self.visit_iHdlExpr(item)
                    if not last:
                        w(", ")
                w(")")
            else:
                assert not sens

        # to prevent useless newline for empty always/time waits
        if skip_body:
            return True
        else:
            return self.visit_iHdlStatement_in_statement(body)

    def visit_iHdlStatement_in_statement(self, stm):
        """
        Print statement which is body of other statement
        e.g. body of process, branch of if-then-else or case of case stememnt
        """
        w = self.out.write
        if isinstance(stm, HdlStmBlock):
            if len(stm.body) == 1 and not stm.labels:
                stm = stm.body[0]
            else:
                w(" ")
                return self.visit_HdlStmBlock(stm)

        w("\n")
        with Indent(self.out):
            return self.visit_iHdlStatement(stm)

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        self.visit_doc(o)
        w = self.out.write
        if o.in_preproc:
            w("generate ")

        w("begin")
        if o.labels:
            w(": ")
            w(o.labels[0])
        w("\n")
        with Indent(self.out):
            for s in o.body:
                need_semi = self.visit_iHdlStatement(s)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
        w("end")
        if o.in_preproc:
            w(" endgenerate")
        return False

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        self.visit_doc(o)
        w = self.out.write
        if o.in_preproc:
            w("generate ")
        w("if (")
        self.visit_iHdlExpr(o.cond)
        w(")")
        need_semi = self.visit_iHdlStatement_in_statement(o.if_true)

        for cond, stms in o.elifs:
            if need_semi:
                w(";\n")
            else:
                w(" ")
            w("else if (")
            self.visit_iHdlExpr(cond)
            w(")")
            need_semi = self.visit_iHdlStatement_in_statement(stms)

        ifFalse = o.if_false
        if ifFalse is not None:
            if need_semi:
                w(";\n")
            else:
                w(" ")
            w("else")
            need_semi = self.visit_iHdlStatement_in_statement(ifFalse)

        if o.in_preproc:
            if need_semi:
                w(";\n")
            else:
                w("\n")
            w("endgenerate")
            return False
        else:
            return need_semi

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        ts = self.top_stm
        if not self.is_in_loop_spec and (ts is o or (ts is not None and ts.in_preproc)):
            w("assign ")
            self.visit_iHdlExpr(o.dst)
            w(" = ")
        else:
            self.visit_iHdlExpr(o.dst)
            if o.is_blocking:
                w(" = ")
            else:
                w(" <= ")

        if o.time_delay is not None:
            w("#")
            self.visit_iHdlExpr(o.time_delay)
            w(" ")
        if o.event_delay is not None:
            w("@")
            if len(o.event_delay) > 1:
                w("(")
            for is_last, e in iter_with_last(o.event_delay):
                self.visit_iHdlExpr(e)
                if not is_last:
                    w(", ")
            if len(o.event_delay) > 1:
                w(")")
            w(" ")

        self.visit_iHdlExpr(o.src)
        return True

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase

        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        if o.uniq_constrain is not None:
            w(o.uniq_constrain.name.lower())
            w(" ")

        w(o.type.name.lower())
        w("(")
        self.visit_iHdlExpr(o.switch_on)
        w(")\n")
        with Indent(self.out):
            cases = o.cases
            for k, stms in cases:
                self.visit_iHdlExpr(k)
                w(":")
                need_semi = self.visit_iHdlStatement_in_statement(stms)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
            defal = o.default
            if defal is not None:
                w("default:")
                need_semi = self.visit_iHdlStatement_in_statement(defal)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
        w("endcase")
        return False

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait

        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        w("#")
        assert len(o.val) == 1, o.val
        self.visit_iHdlExpr(o.val[0])
        return True

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor

        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        if o.in_preproc:
            w("generate ")

        w("for (")
        if isinstance(o.init, HdlStmBlock):
            init_stms = o.init.body
        else:
            init_stms = [o.init, ]

        trnt = self._type_requires_nettype
        in_in_loop_spec = self.is_in_loop_spec
        try:
            self._type_requires_nettype = False
            self.is_in_loop_spec = True
            for is_last, stm in iter_with_last(init_stms):
                self.visit_iHdlStatement(stm)
                if not is_last:
                    w(", ")
        finally:
            self.is_in_loop_spec = in_in_loop_spec
            self._type_requires_nettype = trnt

        w("; ")
        self.visit_iHdlExpr(o.cond)
        w("; ")
        if isinstance(o.step, HdlStmBlock):
            step_stms = o.step.body
        else:
            step_stms = [o.step, ]

        for is_last, stm in iter_with_last(step_stms):
            self.visit_iHdlStatement(stm)
            if not is_last:
                w(", ")
        w(")")
        need_semi = self.visit_iHdlStatement_in_statement(o.body)
        if o.in_preproc:
            if need_semi:
                w(";\n")
            else:
                w("\n")
            w("endgenerate")
            return False
        else:
            return need_semi

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        raise NotImplementedError()

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        w("while (")
        self.visit_iHdlExpr(o.cond)
        w(") ")
        return self.visit_iHdlStatement(o.body)

    def visit_HdlStmRepeat(self, o):
        """
        :type o: HdlStmRepeat
        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        w("repeat (")
        self.visit_iHdlExpr(o.n)
        w(") ")
        return self.visit_iHdlStatement(o.body)

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        self.visit_doc(o)
        w = self.out.write
        w("return")
        if o.val is not None:
            w(" ")
            self.visit_iHdlExpr(o.val)
        return True

    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        self.visit_doc(o)
        self.out.write("continue")
        return True

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        self.visit_doc(o)
        self.out.write("break")
        return True

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        raise ValueError("SystemVerilog does not have exceptions")

    # [TODO] SV only
    def visit_HdlImport(self, o):
        """
        :type o: HdlImport
        """
        self.visit_doc(o)
        w = self.out.write
        w("import ")
        package, name = o.path
        self.visit_iHdlExpr(package)
        w("::")
        self.visit_iHdlExpr(name)

        return True



from itertools import chain

from hdlConvertorAst.hdlAst import HdlImport, HdlStmProcess, HdlStmIf, \
    HdlStmAssign, HdlStmCase, HdlStmWait, HdlStmReturn, HdlStmFor, HdlStmForIn, \
    HdlStmWhile, HdlStmBlock, iHdlStatement, HdlModuleDec, HdlModuleDef, \
    HdlValueIdspace, HdlIdDef, HdlFunctionDef, HdlOp, HdlCompInst, \
    HdlValueInt, HdlStmBreak, HdlStmContinue, HdlStmRepeat, HdlLibrary, HdlContext, \
    HdlClassDef, HdlPhysicalDef, HdlEnumDef, ALL_STATEMENT_CLASSES, HdlStmNop,\
    HdlValueId


class HdlAstVisitor(object):
    """
    A visitor which can be used to traverse AST (Abstract Syntax Tree) made of objects from `hdlConvertorAst.hdlAst` module.

    """

    def __init__(self):
        self._visit_call_dispatch_dict = {
            cls: getattr(self, "visit_" + cls.__name__)
            for cls in [
                HdlContext, HdlImport, HdlLibrary, HdlModuleDec, HdlModuleDef,
                HdlValueIdspace, HdlIdDef, HdlFunctionDef,
                HdlClassDef, HdlPhysicalDef, HdlEnumDef,
                HdlCompInst, HdlOp, HdlValueInt] + list(ALL_STATEMENT_CLASSES)
        }

    def visit_iHdlObj(self, o):
        visit_fn = self._visit_call_dispatch_dict.get(o.__class__, None)
        if visit_fn is not None:
            return visit_fn(o)
        else:
            return self.visit_iHdlExpr(o)

    def visit_doc(self, o):
        pass

    def visit_HdlContext(self, context):
        """
        :type context: HdlContext
        """
        for o in context.objs:
            self.visit_main_obj(o)

        return context

    def visit_HdlImport(self, o):
        """
        :type o: HdlImport
        """
        return o

    def visit_HdlLibrary(self, o):
        """
        :type o: HdlLibrary
        """
        return o

    def visit_HdlValueIdspace(self, o):
        """
        :type o: HdlValueIdspace
        """
        self.visit_doc(o)
        for o2 in o.objs:
            self.visit_iHdlObj(o2)
        return o

    def visit_main_obj(self, o):
        visit_fn = self._visit_call_dispatch_dict.get(o.__class__, None)
        if visit_fn is not None:
            return visit_fn(o)
        else:
            return self.visit_iHdlExpr(o)

    def visit_iHdlStatement(self, o):
        """
        :type o: iHdlStatement
        """
        # statement can be also expressin
        visit_fn = self._visit_call_dispatch_dict.get(o.__class__, None)
        if visit_fn is not None:
            return visit_fn(o)
        elif o is None:
            return self.visit_HdlStmNop(HdlStmNop())
        else:
            return self.visit_iHdlExpr(o)

    def visit_HdlModuleDec(self, o):
        """
        :type o: HdlModuleDec
        """
        self.visit_doc(o)
        for p in o.params:
            self.visit_param(p)
        for p in o.ports:
            self.visit_port(p)
        for o2 in o.objs:
            self.visit_main_obj(o2)
        return o

    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        self.visit_doc(o)
        self.visit_type(o.type)
        if o.value is not None:
            self.visit_iHdlExpr(o.value)
        return o

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        if isinstance(o, HdlOp):
            return self.visit_HdlOp(o)
        return o

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        :return: iHdlExpr
        """
        for op in o.ops:
            self.visit_iHdlExpr(op)
        return o

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        return o

    def visit_port(self, o):
        """
        :type o: HdlIdDef
        """
        return self.visit_HdlIdDef(o)

    def visit_param(self, o):
        """
        :type o: HdlIdDef
        """
        return self.visit_HdlIdDef(o)

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDef
        """
        if o.dec is not None:
            self.visit_HdlModuleDec(o.dec)

        self.visit_doc(o)
        for _o in o.objs:
            if isinstance(_o, iHdlStatement):
                self.visit_iHdlStatement(_o)
            elif isinstance(_o, HdlIdDef):
                self.visit_HdlIdDef(_o)
            elif isinstance(_o, HdlCompInst):
                self.visit_HdlCompInst(_o)
            elif isinstance(_o, HdlFunctionDef):
                self.visit_HdlFunctionDef(_o)
            elif isinstance(_o, HdlModuleDec):
                # vhdl components
                self.visit_HdlModuleDec(_o)
            elif isinstance(_o, HdlOp):
                assert _o.ops[0] == HdlValueId("assert"), _o
                self.visit_HdlOp(_o)
            else:
                raise NotImplementedError(_o)
        return o

    def visit_HdlCompInst(self, o):
        """
        :type o: HdlCompInst
        """
        self.visit_doc(o)
        for pm in chain(o.param_map, o.port_map):
            self.visit_iHdlExpr(pm)
        return o

    def visit_HdlClassDef(self, o):
        """
        :type o: HdlClassDef
        """
        self.visit_doc(o)
        for t in o.base_types:
            self.visit_iHdlExpr(t)

        for m in o.members:
            self.visit_iHdlObj(m)

        return o

    def visit_HdlPhysicalDef(self, o):
        """
        :type o: HdlPhysicalDef
        """
        return o

    def visit_HdlEnumDef(self, o):
        """
        :type o: HdlEnumDef
        """
        return o

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        self.visit_doc(o)
        for p in o.params:
            self.visit_HdlIdDef(p)
        if o.return_t is not None:
            self.visit_iHdlExpr(o.return_t)
        for o2 in o.body:
            self.visit_main_obj(o2)
        return o

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        self.visit_doc(o)
        self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        self.visit_doc(o)
        for o2 in o.body:
            self.visit_iHdlStatement(o2)
        return o

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.switch_on)
        for c, stm in o.cases:
            self.visit_iHdlExpr(c)
            self.visit_iHdlStatement(stm)
        if o.default is not None:
            self.visit_iHdlStatement(o.default)
        return o

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.val)
        return o

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.cond)
        if o.if_true is not None:
            self.visit_iHdlStatement(o.if_true)
        for c, stm in o.elifs:
            self.visit_iHdlExpr(c)
            self.visit_iHdlStatement(stm)
        if o.if_false is not None:
            self.visit_iHdlStatement(o.if_false)
        return o

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        self.visit_doc(o)
        self.visit_iHdlStatement(o.init)
        self.visit_iHdlExpr(o.cond)
        self.visit_iHdlStatement(o.step)
        self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        """
        self.visit_doc(o)
        for v in o.var_defs:
            self.visit_main_obj(v)
        self.visit_iHdlExpr(o.collection)
        self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.cond)
        self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmRepeat(self, o):
        """
        :type o: HdlStmRepeat
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.n)
        self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        self.visit_doc(o)
        if o.val is not None:
            self.visit_iHdlExpr(o.val)
        return o

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        self.visit_doc(o)
        return o

    def visit_HdlStmNop(self, o):
        """
        :type o: HdlStmNop
        """
        self.visit_doc(o)
        return o

    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        self.visit_doc(o)
        return o

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.val)
        return o

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        self.visit_doc(o)
        self.visit_iHdlExpr(o.src)
        self.visit_iHdlExpr(o.dst)
        if o.event_delay is not None:
            self.visit_iHdlExpr(o.event_delay)
        if o.time_delay is not None:
            self.visit_iHdlExpr(o.time_delay)
        return o

    def visit_type(self, t):
        """
        :type t: iHdlExpr
        """
        return self.visit_iHdlExpr(t)



from typing import List, Optional


try:
    # python2
    from StringIO import StringIO
except ImportError:
    # python3
    from io import StringIO


class iHdlObj(object):
    """
    Object with direct representation in HDL

    :ivar ~.doc: doc from the HDL related probably to this object
    :ivar ~.origin: an object which is responsible for this HdlObject
        (not used by this library, used in user code)
    """
    __slots__ = ["doc", "position", "origin"]

    def __init__(self):
        self.doc = ""  # type: str
        self.position = None  # type: Optional[CodePosition]
        self.origin = None

    def __repr__(self):
        from hdlConvertorAst.to.json_debug import ToJsonDebug
        from pprint import pprint
        to = ToJsonDebug()
        d = getattr(to, "visit_" + self.__class__.__name__)(self)
        s = StringIO()
        pprint(d, stream=s, depth=None)
        return s.getvalue()


class iHdlObjWithName(iHdlObj):
    """
    :ivar ~.name: reference to name of the object
    """
    __slots__ = ["name"]

    def __init__(self):
        super(iHdlObjWithName, self).__init__()
        self.name = None  # type: HdlValueId


class iHdlObjInModule(iHdlObj):
    """
    Object which can appear in the module body
    """
    __slots__ = []


class iHdlStatement(iHdlObjInModule):
    """
    :ivar ~.labels: list of labels, the first label is for this statement
        the others are for it's branches
    :ivar ~.in_preproc: if True the statement is VHDL generate
            or other statement which should be evaluated
            compile time (note that this corresponds to VHDL generate statements
            and verilog generate not to a Verilog preprocessor ifdefs
            as they are processed before Verilog code parsing
            and are evaluated by preprocessor
            in the first place)
    """
    __slots__ = ["labels", "in_preproc"]

    def __init__(self):
        iHdlObjInModule.__init__(self)
        self.labels = []  # type: List[HdlValueId]
        self.in_preproc = False  # type: bool



from hdlConvertorAst.hdlAst import HdlOp, HdlValueId, HdlFunctionDef, HdlOpType
from hdlConvertorAst.to.hdl_ast_modifier import HdlAstModifier
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call


class AddCallOperatorForCallWithoutParenthesis(HdlAstModifier):
    """
    Verilog function call does not need to have () and it can be called just by its id.
    To simplify handling we decorete each such a call with a call operator in this transformation.
    """

    def __init__(self):
        HdlAstModifier.__init__(self)
        self._parentExpr = None

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        if isinstance(o, HdlOp):
            prev_par_expr = self._parentExpr
            self._parentExpr = o
            try:
                self.visit_HdlOp(o)
            finally:
                self._parentExpr = prev_par_expr
        else:
            if isinstance(o, HdlValueId) and\
                    isinstance(o.obj, HdlFunctionDef) and \
                    ( not isinstance(self._parentExpr, HdlOp) or \
                      self._parentExpr.fn != HdlOpType.CALL or \
                      self._parentExpr.ops[0] is not o
                      ):
                # wrap function id in a call operator if parent is not a call operator
                return hdl_call(o, [])
        return o



from itertools import chain

from hdlConvertorAst.hdlAst import HdlIdDef, HdlOp, HdlOpType, \
    HdlDirection, HdlValueId, HdlCompInst, HdlPhysicalDef, HdlEnumDef, HdlClassDef, \
    HdlFunctionDef, iHdlStatement
from hdlConvertorAst.hdlAst._expr import HdlTypeType
from hdlConvertorAst.hdlAst._statements import ALL_STATEMENT_CLASSES
from hdlConvertorAst.to.basic_hdl_sim_model.stm import ToBasicHdlSimModelStm
from hdlConvertorAst.to.basic_hdl_sim_model.utils import sensitivityByOp
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last

DEFAULT_IMPORTS = """\
from pyMathBitPrecise.array3t import Array3t, Array3val
from pyMathBitPrecise.bits3t import Bits3t, Bits3val
from pyMathBitPrecise.enum3t import define_Enum3t

from hwtSimApi.basic_hdl_simulator.model import BasicRtlSimModel
from hwtSimApi.basic_hdl_simulator.model_utils import sensitivity, connectSimPort
from hwtSimApi.basic_hdl_simulator.proxy import BasicRtlSimProxy
from hwtSimApi.basic_hdl_simulator.sim_utils import sim_eval_cond
"""


class ToBasicHdlSimModel(ToBasicHdlSimModelStm):
    """
    Convert hdlObject AST to BasicHdlSimModel
    (Python simulation model used by hwtSimApi simulator)

    :ivar ~.module_path_prefix: if None serialized output will be in a single file
        and no imports of components are produced, else
    :type ~.module_path_prefix: Optional[str]
    :type stm_outputs: Optional[Dict[iHdlStm, List[str]]]
    :attention: if you want to serialize a module definitions you need to set stm_outputs
    """
    ALL_STATEMENT_CLASSES = ALL_STATEMENT_CLASSES

    def __init__(self, out_stream):
        """
        :type out_stream: StringIO
        """
        ToBasicHdlSimModelStm.__init__(self, out_stream)
        self.module_path_prefix = None
        self.add_imports = True
        self.stm_outputs = None

    def visit_doc(self, obj):
        return super(ToBasicHdlSimModel, self).visit_doc(obj, "#")

    def visit_component_imports(self, components):
        w = self.out.write
        prefix = self.module_path_prefix
        if prefix is not None:
            for c in components:
                n = c.module_name
                w('from %s.%s import %s\n' % (prefix, n, n))

    def split_HdlModuleDefObjs(self, objs):
        """
        :type mod: List[iHdlObj]
        """
        definitions = []
        variables = []
        processes = []
        components = []
        others = []
        obj_type_containers = {
            HdlClassDef: definitions,
            HdlPhysicalDef: definitions,
            HdlEnumDef: definitions,
            HdlFunctionDef: definitions,
            HdlIdDef: variables,
            HdlCompInst: components,
            HdlOp: others,
        }

        for stmCls in self.ALL_STATEMENT_CLASSES:
            obj_type_containers[stmCls] = processes

        for o in objs:
            if o.__class__ is HdlIdDef and o.type == HdlTypeType:
                definitions.append(o)
            else:
                if isinstance(o, iHdlStatement) and o.in_preproc:
                    others.append(o)
                else:
                    obj_type_containers[o.__class__].append(o)

        return definitions, variables, processes, components, others

    def visit_HdlModuleDef(self, mod_def):
        """
        :type mod_def: HdlModuleDef
        """
        mod_dec = mod_def.dec
        assert mod_dec is not None
        w = self.out.write
        if self.add_imports:
            w(DEFAULT_IMPORTS)
            w("\n")
            if self.module_path_prefix is None:
                self.add_imports = False

        types, variables, processes, components, others = self.split_HdlModuleDefObjs(mod_def.objs)

        self.visit_component_imports(components)
        self.visit_doc(mod_dec)
        w("class ")
        w(mod_dec.name)
        w("(BasicRtlSimModel):\n")
        assert not mod_dec.params, "generic should be already resolved for this format"

        with Indent(self.out):
            for t in types:
                self.visit_type_declr(t)

            w('def __init__(self, sim: "BasicRtlSimulator", name="')
            w(mod_dec.name)
            w('"):\n')
            with Indent(self.out):
                w('BasicRtlSimModel.__init__(self, sim, name=name)\n')
                w('# ports\n')
                for port in mod_dec.ports:
                    self.visit_HdlIdDef(port)
                w("# internal signals\n")
                for v in variables:
                    self.visit_HdlIdDef(v)
                w("# component instances\n")
                for c in components:
                    w("self.")
                    w(c.name.val)
                    w(" = ")
                    w(c.module_name.val)
                    w('(sim, "')
                    w(c.name.val)
                    w('")\n')

                if others:
                    w("# others")
                    for o in others:
                        self.visit_iHdlObj(o)

            w("def _init_body(self):\n")
            with Indent(self.out):
                for c in components:
                    for pm in c.port_map:
                        w("connectSimPort(self, self.")
                        w(c.name.val)
                        w(', "')
                        assert isinstance(pm, HdlOp) and\
                            pm.fn == HdlOpType.MAP_ASSOCIATION, pm
                        mod_port, connected_sig = pm.ops
                        assert isinstance(
                            connected_sig, HdlValueId), connected_sig
                        self.visit_iHdlExpr(connected_sig)
                        w('", "')
                        assert isinstance(mod_port, HdlValueId), mod_port
                        self.visit_iHdlExpr(mod_port)
                        w('")\n')

                w('self._interfaces = (\n')
                with Indent(self.out):
                    for p in chain(mod_dec.ports, variables):
                        if not p.is_const:
                            w("self.io.")
                            w(p.name)
                            w(',\n')
                w(')\n')

                w('self._processes = (\n')
                with Indent(self.out):
                    for p in processes:
                        w("self.")
                        try:
                            w(p.labels[0])
                        except Exception:
                            raise
                        w(",\n")
                w(")\n")
                w('self._units = (')
                with Indent(self.out):
                    for c in components:
                        w("self.")
                        w(c.name.val)
                        w(",\n")
                w(")\n")

                for proc in processes:
                    w("sensitivity(self.")
                    w(proc.labels[0])
                    w(', ')
                    for last, s in iter_with_last(proc.sensitivity):
                        if isinstance(s, HdlOp):
                            w("(")
                            w(str(sensitivityByOp(s.fn)))
                            w(", self.io.")
                            self.visit_iHdlExpr(s.ops[0])
                            w(')')
                        else:
                            w("self.io.")
                            self.visit_iHdlExpr(s)
                        if not last:
                            w(", ")
                    w(")\n")
                    w("self._outputs[self.")
                    w(proc.labels[0])
                    w("] = (\n")
                    outputs = self.stm_outputs[proc]
                    with Indent(self.out):
                        for outp in outputs:
                            w("self.io.")
                            assert isinstance(outp, HdlValueId)
                            w(outp.val)
                            w(",\n")
                    w(")\n")
                w("for u in self._units:\n")
                w("    u._init_body()\n\n")
            for p in processes:
                self.visit_HdlStmProcess(p)
                # extra line to separate a process functions
                w("\n")

    def visit_type_declr(self, t):
        """
        :type t: HdlIdDef
        """
        self.visit_doc(t)
        w = self.out.write
        w(t.name)
        w(" = ")
        val = t.value
        if isinstance(val, HdlPhysicalDef):
            # I don't get what this is, TODO
            raise NotImplementedError("HdlPhysicalDef is a VHDL specific type"
                                      " and does not have direct equivalet in this format")
        elif isinstance(val, HdlEnumDef):
            w('define_Enum3t("')
            w(t.name)
            w('", [')
            for last, (k, v) in iter_with_last(val.values):
                assert v is None, v
                w('"%s"' % k)
                if not last:
                    w(", ")
            w("])()\n")
        elif isinstance(val, HdlOp) and val.fn == HdlOpType.INDEX:
            # array type def.
            self.visit_iHdlExpr(val)
            w("\n")
        else:
            raise NotImplementedError()

    def visit_HdlIdDef(self, var):
        """
        :type var: HdlIdDef
        """
        self.visit_doc(var)
        w = self.out.write
        if var.is_const:
            w("self.")
            w(var.name)
            w(" = ")
            self.visit_iHdlExpr(var.value)
            w("\n")
        else:
            w("self.io.")
            w(var.name)
            w(' = BasicRtlSimProxy(\n')
            with Indent(self.out):
                w('sim, self, "')
                w(var.name)
                w('",\n')
                self.visit_type(var.type)
                w(", ")
            if var.value is None:
                w("None)\n")
            else:
                self.visit_iHdlExpr(var.value)
                w(")\n")

    def visit_HdlContext(self, context, stm_outputs):
        """
        :type context: HdlContext
        :type stm_outputs: Dict[HdlStm, List[HdlValueId]]
        """
        self.stm_outputs = stm_outputs
        return super(ToBasicHdlSimModel, self).visit_HdlContext(context)



import keyword
import sys


HWT_KEYWORDS = keyword.kwlist + [
    # imports
    'hwt', 'Unit', "Param", "Signal",
    "Concat", "If", "Switch",
    "HArray", "Bits", "INT", "SLICE", "STR", "BIT", "FLOAT64",
    "HEnum", "_", "rename_signal",
]

if sys.version_info[0] <= 2:
    HWT_KEYWORDS.append("None")



from hdlConvertorAst.hdlAst import HdlOpType, HdlValueId, HdlValueInt, \
    HdlOp
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.common import ToHdlCommon, ASSOCIATIVITY,\
    ASSIGN_OPERATORS_SYMBOLS_C
from hdlConvertorAst.to.hdlUtils import iter_with_last
from hdlConvertorAst.to.verilog.expr import ASSIGN_OPERATORS, ToVerilog2005Expr


L = ASSOCIATIVITY.L_TO_R
R = ASSOCIATIVITY.R_TO_L


class ToSystemcExpr(ToHdlCommon):
    # https://en.cppreference.com/w/cpp/language/operator_precedence
    OP_PRECEDENCE = {
        HdlOpType.DOUBLE_COLON: (1, L),

        HdlOpType.INCR_POST: (2, L),
        HdlOpType.DECR_POST: (2, L),
        HdlOpType.TYPE_OF: (2, L),
        HdlOpType.CALL: (2, L),
        HdlOpType.INDEX: (3, L),
        HdlOpType.DOT: (2, L),
        HdlOpType.ARROW: (2, L),
        HdlOpType.PARAMETRIZATION: (2, L),


        HdlOpType.INCR_PRE: (3, R),
        HdlOpType.DECR_PRE: (3, R),
        HdlOpType.MINUS_UNARY: (3, R),
        HdlOpType.PLUS_UNARY: (3, R),
        HdlOpType.NEG_LOG: (3, R),
        HdlOpType.NEG: (3, R),
        HdlOpType.REFERENCE: (3, R),

        HdlOpType.DIV: (5, L),
        HdlOpType.MUL: (5, L),
        HdlOpType.MOD: (5, L),

        HdlOpType.ADD: (6, L),
        HdlOpType.SUB: (6, L),

        HdlOpType.SLL: (7, L),
        HdlOpType.SRL: (7, L),

        HdlOpType.GT:  (9, L),
        HdlOpType.LT:  (9, L),
        HdlOpType.GE:  (9, L),
        HdlOpType.LE:  (9, L),

        HdlOpType.NE: (10, L),
        HdlOpType.EQ: (10, L),

        HdlOpType.AND: (11, L),
        HdlOpType.XOR: (12, L),
        HdlOpType.OR: (13, L),

        HdlOpType.AND_LOG: (14, L),
        HdlOpType.OR_LOG: (15, L),

        HdlOpType.TERNARY: (16, R),
    }
    OP_PRECEDENCE.update({
        k: (16, R) for k in ASSIGN_OPERATORS
    })

    GENERIC_UNARY_OPS = {
        HdlOpType.NEG_LOG: "!",
        HdlOpType.NEG: "~",
        HdlOpType.MINUS_UNARY: "-",
        HdlOpType.PLUS_UNARY: "+",
        HdlOpType.INCR_PRE: "++",
        HdlOpType.DECR_PRE: "--",
    }
    GENERIC_UNARY_OPS_POSTFIX = ToVerilog2005Expr.GENERIC_UNARY_OPS_POSTFIX

    GENERIC_BIN_OPS = {
        HdlOpType.DOUBLE_COLON: "::",
        HdlOpType.AND: " & ",
        HdlOpType.OR: " | ",
        HdlOpType.XOR: " ^ ",

        HdlOpType.EQ: ' == ',
        HdlOpType.NE: " != ",

        HdlOpType.MUL: " * ",
        HdlOpType.DIV: " / ",
        HdlOpType.POW: " ** ",
        HdlOpType.MOD: " % ",

        HdlOpType.SLL: " << ",
        HdlOpType.SRL: " >> ",
    }
    GENERIC_BIN_OPS.update(ToHdlCommon.GENERIC_BIN_OPS)
    GENERIC_BIN_OPS.update(ASSIGN_OPERATORS_SYMBOLS_C)

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        w = self.out.write
        if o.bits is None:
            w(str(o.val))
        else:
            if o.base is None:
                w(str(o.val))
            else:
                b = o.base
                if isinstance(o.val, int):
                    if b == 2:
                        f = "0b{0:b}"
                    elif b == 8:
                        f = "0o{0:o}"
                    elif b == 16:
                        f = "0x{0:x}"
                    else:
                        raise NotImplementedError(b)
                    w(f.format(o.val))
                else:
                    if b == 2:
                        f = '"0b{0}"'
                    elif b == 8:
                        f = '"0o{0}"'
                    elif b == 16:
                        f = '"0x{0}"'
                    else:
                        raise NotImplementedError(b)

                    w(f.format(o.val.upper()))

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        """
        w = self.out.write
        if isinstance(o, HdlValueId):
            w(o.val)
        elif is_str(o):
            w('"%s"' % o)
        elif isinstance(o, HdlValueInt):
            self.visit_HdlValueInt(o)
        elif isinstance(o, (list, tuple)):
            with_nl = len(o) > 3
            if isinstance(o, list):
                begin = "{"
                end = "}"
            else:
                begin = "("
                end = ")"

            w(begin)
            for elem in o:
                self.visit_iHdlExpr(elem)
                if with_nl:
                    w(", \n")
                else:
                    w(", ")
            w(end)
        elif isinstance(o, HdlOp):
            self.visit_HdlOp(o)
        elif o is None:
            w("nullptr")
        else:
            raise NotImplementedError(o.__class__, o)
        return True

    def visit_HdlOp(self, op):
        """
        :type o: HdlOp
        """

        w = self.out.write
        o = op.fn
        if o == HdlOpType.RISING:
            self._visit_operand(op.ops[0], 0, op, False, False)
            w(".pos()")
        elif o == HdlOpType.FALLING:
            self._visit_operand(op.ops[0], 0, op, False, False)
            w(".neg()")
        elif o == HdlOpType.PARAMETRIZATION:
            self._visit_operand(op.ops[0], 0, op, False, False)
            w("<")
            for last, _o in iter_with_last(op.ops[1:]):
                self.visit_iHdlExpr(_o)
                if not last:
                    w(", ")
            w(">")
        else:
            ToHdlCommon.visit_HdlOp(self, op)
        return True



from hdlConvertorAst.hdlAst import iHdlStatement, HdlStmCaseType, HdlIdDef, \
    HdlOp, HdlOpType, HdlStmBlock
from hdlConvertorAst.to.hdlUtils import iter_with_last, Indent, UnIndent
from hdlConvertorAst.to.vhdl.expr import ToVhdl2008Expr


class ToVhdl2008Stm(ToVhdl2008Expr):

    def visit_assert(self, args):
        """
        :type args: List[iHdlExpr]
        """
        w = self.out.write
        w("ASSERT ")
        for is_last, (prefix, a) in iter_with_last(
                zip(("", "REPORT ", "SEVERITY "), args)):
            w(prefix)
            self.visit_iHdlExpr(a)
            if not is_last:
                w(" ")

    def visit_report(self, args):
        """
        :type args: List[iHdlExpr]
        """
        w = self.out.write
        w("REPORT ")
        for is_last, (prefix, a) in iter_with_last(
                zip(("", "SEVERITY "), args)):
            w(prefix)
            self.visit_iHdlExpr(a)
            if not is_last:
                w(" ")

    def visit_HdlStmProcess(self, o):
        """
        :type proc: HdlStmProcess
        """
        self.visit_doc(o)
        sens = o.sensitivity
        w = self.out.write
        if o.labels:
            w(o.labels[0])
            w(": ")
        w("PROCESS")
        if o.trigger_constrain is not None:
            raise NotImplementedError()
        if sens:
            w("(")
            for last, item in iter_with_last(sens):
                self.visit_iHdlExpr(item)
                if not last:
                    w(", ")
            w(")")
        w("\n")
        self.visit_HdlStmBlock(o.body, force_space_before=False, force_begin_end=True)
        w(" PROCESS;\n")

    def _write_begin(self, begin_end, must_have_begin_end, force_space_before):
        w = self.out.write
        if begin_end and must_have_begin_end:
            if force_space_before:
                w(" BEGIN\n")
            else:
                w("BEGIN\n")
        else:
            w("\n")

    def visit_HdlStmBlock(self, stms, force_space_before=True, begin_end=True, force_begin_end=False):
        """
        :type stms: Union[List[iHdlStatement], iHdlStatement, iHdlExpr]
        :return: True if statements are wrapped in begin-end block
        """
        w = self.out.write
        in_preproc = False
        if isinstance(stms, HdlStmBlock):
            self.visit_doc(stms)
            must_have_begin_end = True
            in_preproc = stms.in_preproc
            stms = stms.body
        elif isinstance(stms, list):
            must_have_begin_end = len(stms) != 1
        else:
            must_have_begin_end = False
            stms = [stms, ]
        must_have_begin_end |= force_begin_end

        non_declarative_seen = False
        with Indent(self.out):
            for s in stms:
                if isinstance(s, iHdlStatement):
                    if not non_declarative_seen:
                        non_declarative_seen = True
                        with UnIndent(self.out):
                            self._write_begin(begin_end, must_have_begin_end, force_space_before)
                    self.visit_iHdlStatement(s)
                elif isinstance(s, HdlIdDef):
                    self.visit_HdlIdDef(s)
                else:
                    if not non_declarative_seen:
                        non_declarative_seen = True
                        with UnIndent(self.out):
                            self._write_begin(begin_end, must_have_begin_end, force_space_before)
                    if in_preproc:
                        self.visit_iHdlObj(s)
                        w("\n")
                    else:
                        self.visit_iHdlExpr(s)
                        w(";\n")

        if not non_declarative_seen:
            self._write_begin(begin_end, must_have_begin_end, force_space_before)

        if begin_end and must_have_begin_end:
            w("END")
            return True

        return False

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        self.visit_doc(o)
        w = self.out.write
        if o.labels:
            w(o.labels[0])
            w(": ")
        w("IF ")
        self.visit_iHdlExpr(o.cond)
        if o.in_preproc:
            w(" GENERATE")
        else:
            w(" THEN")
        need_space = self.visit_HdlStmBlock(o.if_true, begin_end=False)

        for cond, stms in o.elifs:
            if need_space:
                w(" ")
            w("ELSIF ")
            self.visit_iHdlExpr(cond)
            if o.in_preproc:
                w(" GENERATE")
            else:
                w(" THEN")
            need_space = self.visit_HdlStmBlock(stms, begin_end=False)

        ifFalse = o.if_false
        if ifFalse is not None:
            if need_space:
                w(" ")
            w("ELSE")
            if o.in_preproc:
                w(" GENERATE")
            self.visit_HdlStmBlock(ifFalse, begin_end=False)
        if need_space:
            w("\n")
        if o.in_preproc:
            w("END GENERATE;\n")
        else:
            w("END IF;\n")

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        w = self.out.write
        if o.time_delay is not None:
            raise NotImplementedError()
        if o.event_delay is not None:
            raise NotImplementedError()

        self.visit_iHdlExpr(o.dst)
        if o.is_blocking:
            w(" := ")
        else:
            w(" <= ")
        with Indent(self.out):
            self.visit_iHdlExpr(o.src)
        w(";\n")

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        self.visit_doc(o)
        w = self.out.write
        if o.type != HdlStmCaseType.CASE:
            raise NotImplementedError()
        if o.uniq_constrain is not None:
            raise NotImplementedError()
        w("CASE ")
        self.visit_iHdlExpr(o.switch_on)
        w(" IS\n")
        with Indent(self.out):
            cases = o.cases
            for k, stms in cases:
                w("WHEN ")
                self.visit_iHdlExpr(k)
                w(" =>")
                is_block = self.visit_HdlStmBlock(stms, begin_end=False)
                if is_block:
                    w("\n")
            defal = o.default
            if defal is not None:
                is_block = w("WHEN OTHERS =>")
                self.visit_HdlStmBlock(defal, begin_end=False)
                if is_block:
                    w("\n")
        w("END CASE;\n")

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        self.visit_doc(o)
        w = self.out.write
        w("RETURN")
        if o.val is not None:
            w(" ")
            self.visit_iHdlExpr(o.val)
        w(";\n")

    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        self.visit_doc(o)
        self.out.write("CONTINUE;\n")

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        self.visit_doc(o)
        self.out.write("BREAK;\n")

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        self.visit_doc(o)
        w = self.out.write
        w("FOR ")
        self.visit_iHdlExpr(o.params[0])
        w(" IN ")
        self.visit_iHdlExpr(o.params[1])
        w(" LOOP\n")
        with Indent(self.out):
            for b in o.body:
                self.visit_iHdlStatement(b)
        w("END FOR;\n")

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        """
        self.visit_doc(o)
        w = self.out.write
        if o.labels:
            w(o.labels[0])
            w(": ")
        w("FOR ")
        assert len(o.var_defs) == 1, o.var_defs
        self.visit_iHdlExpr(o.var_defs[0])
        w(" IN ")
        self.visit_iHdlExpr(o.collection)
        if o.in_preproc:
            w(" GENERATE\n")
        else:
            w(" LOOP\n")
        with Indent(self.out):
            if o.in_preproc:
                has_begin_end = self.visit_iHdlObj(o.body)
            else:
                if isinstance(o.body, HdlStmBlock):
                    for _stm in o.body.body:
                        self.visit_iHdlObj(_stm)
                else:
                    self.visit_iHdlObj(o.body)

        if o.in_preproc:
            if has_begin_end:
                w(" GENERATE;\n")
            else:
                w("\n")
                w("END GENERATE;\n")
        else:
            w("END LOOP;\n")

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        self.visit_doc(o)
        w = self.out.write
        w("WAIT")
        for e in o.val:
            if isinstance(e, HdlOp) and e.fn == HdlOpType.MUL:
                # wait for time
                w(" FOR ")
                self.visit_iHdlExpr(e.ops[0])
                w(" ")
                self.visit_iHdlExpr(e.ops[1])
            # elif :
            #  wait until
            else:
                # wait on event
                w(" ON ")
                self.visit_iHdlExpr(e)

        w(";\n")

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        raise ValueError("VHDL does not have exceptions")

    def visit_HdlStmNop(self, o):
        """
        :type o: HdlStmNop
        """
        self.visit_doc(o)
        self.out.write("NULL;\n")

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        :note: vhdl loop statement
        """
        self.visit_doc(o)
        w = self.out.write
        if o.labels:
            w(o.labels[0])
            w(": ")
        w("LOOP\n")
        with Indent(self.out):
            if not o.in_preproc and isinstance(o.body, HdlStmBlock):
                for _stm in o.body.body:
                    self.visit_iHdlObj(_stm)
            else:
                self.visit_iHdlObj(o.body)
        w("END LOOP;\n")




from copy import deepcopy

from hdlConvertorAst.hdlAst import HdlOpType, HdlOp, HdlValueId, HdlStmThrow
from hdlConvertorAst.to.hdl_ast_modifier import HdlAstModifier
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import \
    to_property_call, hdl_call, hdl_getattr, hdl_add_int, hdl_sub_int


class BasicHdlSimModelTranslateVerilogOperands(HdlAstModifier):

    def __init__(self, downto_to_slice_fn=True):
        super(BasicHdlSimModelTranslateVerilogOperands, self).__init__()
        self.downto_to_slice_fn = downto_to_slice_fn

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        """
        o = HdlAstModifier.visit_HdlOp(self, o)
        op = o.fn
        if op == HdlOpType.EQ or op == HdlOpType.XNOR:
            # HdlOpType.EQ: '%s._eq(%s)',
            to_property_call(o, "_eq")
        elif op == HdlOpType.CONCAT:
            to_property_call(o, "_concat")
        elif op == HdlOpType.TERNARY:
            to_property_call(o, "_ternary__val")
        elif op == HdlOpType.DOWNTO:
            if self.downto_to_slice_fn:
                o.fn = HdlOpType.CALL
                o.ops = [HdlValueId("slice"), hdl_add_int(o.ops[0], 1), o.ops[1]]
            else:
                o.ops = [hdl_add_int(o.ops[0], 1), o.ops[1]]
        elif op == HdlOpType.TO:
            raise NotImplementedError(o)
        elif op == HdlOpType.CALL:
            fn = o.ops[0]
            if fn == HdlValueId("$display"):
                o.ops = [HdlValueId("print"), ] + o.ops[1:]
            elif fn == HdlValueId("$finish"):
                t = HdlStmThrow()
                t.val = hdl_call(HdlValueId("Exception"), [])
                return t
            elif fn == HdlValueId("$time"):
                return hdl_getattr(HdlValueId("sim"), "now")
            elif fn == HdlValueId("$rtoi"):
                o.ops = [HdlValueId("int"), ] + o.ops[1:]
            else:
                return hdl_call(hdl_getattr(HdlValueId("self"), fn.val), o.ops[1:])

        elif op == HdlOpType.REPL_CONCAT:
            n, v = o.ops
            return hdl_call(HdlValueId("replicate"), [n, v])
        elif op in (HdlOpType.AND_LOG, HdlOpType.OR_LOG, HdlOpType.NEG_LOG):
            ops = [hdl_call(hdl_getattr(_o, "_isOn"), []) for _o in o.ops]
            if op == HdlOpType.AND_LOG:
                new_o_fn = HdlOpType.AND
            elif op == HdlOpType.OR_LOG:
                new_o_fn = HdlOpType.OR
            else:
                assert op == HdlOpType.NEG_LOG
                new_o_fn = HdlOpType.NEG
            return HdlOp(new_o_fn, ops)
        elif op == HdlOpType.PART_SELECT_POST:
            # logic [31: 0] a;
            # a[ 0 +: 8] == a[ 7 : 0];
            # logic [0 :31] b;
            # b[ 0 +: 8] == b[0 : 7]
            high, width = o.ops
            o.fn = HdlOpType.CALL
            o.ops = [HdlValueId("slice"),
                    high,
                    HdlOp(HdlOpType.SUB, [deepcopy(high), width])]
        return o



from copy import copy

from hdlConvertorAst.hdlAst import HdlOpType
from hdlConvertorAst.to.basic_hdl_sim_model.expr import ToBasicHdlSimModelExpr
from hdlConvertorAst.to.common import ASSOCIATIVITY


class ToHwtExpr(ToBasicHdlSimModelExpr):
    OP_PRECEDENCE = copy(ToBasicHdlSimModelExpr.OP_PRECEDENCE)
    OP_PRECEDENCE.update({
        HdlOpType.DOWNTO: (1, ASSOCIATIVITY.L_TO_R),
    })
    GENERIC_BIN_OPS = copy(ToBasicHdlSimModelExpr.GENERIC_BIN_OPS)
    GENERIC_BIN_OPS.update({
        HdlOpType.DOWNTO: ":",
    })

    def visit_HdlOp(self, o):
        """
        :type op: HdlOp
        """
        op = o.fn
        if op in (HdlOpType.AND_UNARY, HdlOpType.OR_UNARY, HdlOpType.XOR_UNARY):
            ops = o.ops
            assert len(ops) == 1, ops
            w = self.out.write
            if HdlOpType.AND_UNARY:
                w("And(*")
            elif op is HdlOpType.OR_UNARY:
                w("Or(*")
            elif op is HdlOpType.XOR_UNARY:
                w("Xor(*(")
            else:
                raise NotImplementedError()
            self.visit_iHdlExpr(ops[0])
            w("))")
            return

        return ToBasicHdlSimModelExpr.visit_HdlOp(self, o)



from enum import Enum
from typing import Optional, Union, List

from hdlConvertorAst.hdlAst._bases import iHdlObj
from hdlConvertorAst.py_ver_compatibility import is_str


class HdlDirection(Enum):
    """
    Direction of the interface
    """
    (
        IN,
        OUT,
        INOUT,
        BUFFER,
        LINKAGE,
        INTERNAL,
        UNKNOWN
    ) = range(7)


class HdlValueId(object):
    """
    String which is id in HDL

    :type ~.val: str
    :ivar ~.obj: an object which corresponds to this name
        (has to be explicitly discovered and is not available imediately
        after parsing)
    """
    __weakref__ = ["obj"]
    __slots__ = ["val", "obj"]

    def __init__(self, val, obj=None):
        assert is_str(val), val
        self.val = val
        self.obj = obj

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.val == other.val

    def __lt__(self, other):
        if not isinstance(other, HdlValueId):
            return False
        return self.val < other.val

    def __hash__(self):
        return hash(self.val)

    def __str__(self):
        return self.val

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.val)


class HdlAll(object):
    """
    Constant which corresponds to VHDL "all" keyword
    or "*" in Verilog sensitivity list
    """
    __slots__ = []

    def __init__(self):
        raise ValueError("This class (%r) is not supposed to be instantiated "
                         "and should be used as a constant instead"
                         % self.__class__)


class HdlOthers(object):
    """
    Constant which corresponds to VHDL "others" keyword
    """
    __slots__ = []

    def __init__(self):
        raise ValueError("This class (%r) is not supposed to be instantiated"
                         % self.__class__)


class HdlTypeType(object):
    """
    Type which means that the object is type of type
    """
    __slots__ = []

    def __init__(self):
        raise ValueError("This class (%r) is not supposed to be instantiated"
                         % self.__class__)


class HdlTypeSubtype(object):
    """
    Type which is used in VHDL subtype definitions as a type of symbol in typedef
    """
    __slots__ = []

    def __init__(self):
        raise ValueError("This class (%r) is not supposed to be instantiated"
                         % self.__class__)


class HdlTypeAuto(object):
    """
    Type which means that the type is automatically resolved
    from the type of the value
    """
    __slots__ = []

    def __init__(self):
        raise ValueError("This class (%r) is not supposed to be instantiated"
                         % self.__class__)


class HdlOpType(Enum):
    """
    The build in functions and operators in HDL languages.
    (Python equivalent of c++ hdlConvertor::hdlAst::HdlOpType)
    """
    (
        # arithmetic
        MINUS_UNARY,
        PLUS_UNARY,
        SUB, # -
        ADD, # +
        DIV, # /
        MUL, # *
        MOD, # % modulo
        REM, # remainder
        POW, # power of
        ABS, # absolute value
        INCR_PRE,  # ++x
        DECR_PRE,  # --x
        INCR_POST,  # x--
        DECR_POST,  # x++
        # bitwise log. operators
        NEG_LOG,  # ~, logical negation "not" in vhdl
        NEG,  # bitwise negation
        AND_LOG,  # "and" in vhdl, &&
        OR_LOG,  # "or" in vhdl, ||
        AND,  # & in vhdl
        OR,  # | in vhdl
        NAND,
        NOR,
        XOR,
        XNOR,
        # SV reduction operators
        OR_UNARY,  # or reduction (|a)
        AND_UNARY,  # and reduction (&a)
        NAND_UNARY,  # nand reduction (~&a)
        NOR_UNARY,  # nor reduction (~|a)
        XOR_UNARY,  # xor reduction (^a)
        XNOR_UNARY,  # and reduction (~^a) or (^~a)
        # shifts
        SLL,  # shift left logical
        SRL,  # shift right logical
        SLA,  # shift left arithmetical
        SRA,  # shift right arithmetical
        ROL,  # rotate left
        ROR,  # rotate right
        # comparison operators
        EQ,  # ==
        NE,  # !=
        IS,  # ===
        IS_NOT, # !==
        LT,  # <
        LE,   # <=
        GT,  # >
        GE,  # >=
        # VHDL-2008 matching ops (the X values are ignored while match)
        EQ_MATCH, # SV ==?
        NE_MATCH, # SV !=?
        LT_MATCH,
        LE_MATCH,
        GT_MATCH,
        GE_MATCH,
        # member accessing
        INDEX,  # array index
        CONCAT,  # concatenation of signals
        REPL_CONCAT,  # replicative concatenation {<N>, <item>}
                      # duplicates and concatenates the item N times
        PART_SELECT_POST, # logic [31: 0] a; logic [0 :31] b;  a[ 0 +: 8] == a[ 7 : 0]; b[ 0 +: 8] == b[0 : 7]
        PART_SELECT_PRE, # a[15 -: 8] == a[15 : 8]; b[15 -: 8] == b[8 :15]
        DOT,  # accessing of property
        DOUBLE_COLON,  # ::, SV accessing class/package static property/type
        APOSTROPHE,  # vhdl attribute access
        ARROW,  # pointer member access, VHDL arrow operator used in type descriptions
        REFERENCE,
        DEREFERENCE,
        
        # assignment operators
        ASSIGN,  # =
        PLUS_ASSIGN,  # +=
        MINUS_ASSIGN,  # -=
        MUL_ASSIGN,  # *=
        DIV_ASSIGN,  # /=
        MOD_ASSIGN,  # %=
        AND_ASSIGN,  # &=
        OR_ASSIGN,  # |=
        XOR_ASSIGN,  # ^=
        SHIFT_LEFT_ASSIGN,  # <<=
        SHIFT_RIGHT_ASSIGN,  # >>=
        ARITH_SHIFT_LEFT_ASSIGN,  # <<<=
        ARITH_SHIFT_RIGHT_ASSIGN,  # >>>=
        
        TERNARY, # cond ? a:b, a if cond else b 
        CALL,  # call of HDL function
        RISING,  # rising edge/posedge event operator
        FALLING,  # falling edge/negedge event operator
        DOWNTO,  # downto for the slice specification
        TO,  # to for the slice specification
        PARAMETRIZATION,  # specification of template arguments
        MAP_ASSOCIATION, # arg=val
        RANGE,  # range used in VHDL type specifications
        THROUGHOUT,  # SV throughout operator 
        DEFINE_RESOLVER,  # used in resolver specification in vhdl subtype definition
        TYPE_OF,  # SV type operator
        UNIT_SPEC, # vhdl unit specification eg. 10 ns
    ) = range(87)
    # note that in verilog bitewise operators can have only one argument


class HdlOp(iHdlObj):
    """
    Container for call of the HDL function in HDL code
    """
    __slots__ = ["fn", "ops"]

    def __init__(self, fn, ops):
        """
        :type fn: Union[HdlOpType, iHdlExpr]
        :type ops: List[iHdlExpr]
        """
        self.fn = fn
        self.ops = ops

    def __lt__(self, other):
        if isinstance(other, HdlValueId):
            return True
        else:
            return (self.fn.value, self.ops) < (other.fn.value, other.ops)

    def __eq__(self, other):
        if not isinstance(other, HdlOp):
            return False
        else:
            return self.fn == other.fn and self.ops == other.ops


class HdlValueInt(iHdlObj):
    """
    Object for representation of int value in in HDL
    (= also for the bitstrings)

    :ivar ~.val: int value or bitstring string
    :ivar ~.bits: number of bits if specified
    :ivar ~.base: base for bitstring
    """
    __slots__ = ["val", "bits", "base"]

    def __init__(self, val, bits, base):
        self.val = val  # type: Union[int, str]
        self.bits = bits  # type: Optional[int]
        self.base = base  # type: Optional[Union[2, 8, 10, 16, 256]]

    def __int__(self):
        return int(self.val)

    def __bool__(self):
        return bool(self.val)

    def __hash__(self):
        return hash((self.val, self.bits, self.base))

    def __lt__(self, other):
        return (self.val, self.bits, self.base) < (other.val, other.bits, other.base)

    def __nonzero__(self):
        return self.__bool__()

    def __eq__(self, other):
        if isinstance(other, HdlValueInt):
            return (self.val == other.val
                    and self.bits == other.bits
                    and self.base == other.base)
        else:
            try:
                return self.val == other
            except Exception:
                return False
            except ValueError:
                return False


class HdlExprNotImplemented(iHdlObj):
    """
    An object which means that the orignal object was not converted
    because such a functionality was not implemented.

    Under normal circumstances should not appear in iHdlExpr.
    """
    pass


# None is equivalent of HDL null
iHdlExpr = Union[HdlValueId, HdlValueInt, float, str,
                 None, List["iHdlExpr"], HdlAll, HdlOp]



"""
.. code-block:: verilog

    // sync reset
    always @(posedge clk) begin
       if(reset)
          reg <= 0;
       else
          reg <= reg_next;
    end
    //-->
    if(clk.rising_edge())
       if(reset)
          reg <= 0;
       else
          reg <= reg_next;
    // async reset
    always @(posedge clk, posedge reset) begin
        if(reset == 1'b1)
           reg <= 0;
        else
           reg <= reg_next;
    end
    -->
    if(reset == 1'b1)
       reg <= 0;
    else if(clk.rising_endge())
       reg <= reg_next;
    // latch
    always @(enable, clk) begin
        if(!clk)
           en_out = enable
    end
    -->
    if(!clk)
       en_out = enable
    // comb. logic
    always @(sel, a, b)
      begin
        f = b;
        if (sel == 1)
          f = a;
      end
    -->
    begin
      f = b;
      if (sel == 1)
        f = a;
    end
    // same with always_*
    always_ff @(negedge clk)
        latch_update_en <= next_latch_update_en;
    always_comb
        gated_clk = clk & latch_update_en;
    always_latch
        if(gated_clk)
             latch_data <= new_data;
"""
from hdlConvertorAst.hdlAst._statements import HdlStmIf
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_or
from hdlConvertorAst.hdlAst import HdlAll


class InjectProcessSensToStatements(HdlAstVisitor):
    def visit_iHdlExpr(self, o):
        pass

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        sens = o.sensitivity
        if o.trigger_constrain is not None:
            raise NotImplementedError()
        if not sens or sens == [HdlAll, ]:
            return

        s = hdl_or(*sens)
        i = HdlStmIf()
        i.cond = s
        i.if_true = o.body
        o.body = i



from itertools import chain

from hdlConvertorAst.hdlAst import iHdlStatement, HdlIdDef, \
    HdlModuleDec, HdlModuleDef, HdlCompInst, HdlFunctionDef
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.translate.common.name_scope import WithNameScope


class BuiltIn(object):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)


class DiscoverDeclarations(HdlAstVisitor):

    def __init__(self, name_scope):
        """
        :type name_scope: NameScope
        """
        super(DiscoverDeclarations, self).__init__()
        self.name_scope = name_scope

    def visit_HdlIdDef(self, o):
        """
        :type name_scope: NameScope
        :type o: HdlIdDef
        """
        self.name_scope.register_name(o.name, o)

    def visit_HdlModuleDec(self, o):
        """
        :type name_scope: NameScope
        :type o: HdlModuleDec
        """
        ns = self.name_scope
        ns.register_name(o.name, o)
        with WithNameScope(self, ns.level_push(o.name)):
            for p in chain(o.params, o.ports):
                self.visit_HdlIdDef(p)

            for o2 in o.objs:
                raise NotImplementedError(o2)

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDef
        """
        if o.dec is not None:
            self.visit_HdlModuleDec(o.dec)

        with WithNameScope(self, self.name_scope.get_child(o.module_name.val)):
            self.discover_declarations(o.objs)

    def visit_HdlCompInst(self, o):
        """
        :type o: HdlCompInst
        """
        if o.name is not None:
            self.name_scope.register_name(o.name, o)
        # name_scope = name_scope.get_object_by_name(o.module_name)

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        ns = self.name_scope
        ns.register_name(o.name, o)
        with WithNameScope(self, ns.level_push(o.name)):
            self.discover_declarations(o.params)
            self.discover_declarations(o.body)
        return o

    def _discover_declarations(self, o):
        if isinstance(o, HdlIdDef):
            self.visit_HdlIdDef(o)
        elif isinstance(o, HdlModuleDec):
            self.visit_HdlModuleDec(o)
        elif isinstance(o, HdlModuleDef):
            self.visit_HdlModuleDef(o)
        elif isinstance(o, iHdlStatement):
            pass
        elif isinstance(o, HdlCompInst):
            self.visit_HdlCompInst(o)
        elif isinstance(o, HdlFunctionDef):
            self.visit_HdlFunctionDef(o)
        else:
            raise NotImplementedError(o)

    def discover_declarations(self, objs):
        for o in objs:
            self._discover_declarations(o)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
from setuptools import find_packages, setup
import sys


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md")) as f:
    long_description = f.read()

deps = ["typing", "future"] if sys.version_info[0] == 2 else None

setup(
    name='hdlConvertorAst',
    version='1.1',
    description='A library of AST nodes for HDL languages (Verilog, VHDL, ...) and transpiler/compiler utilities',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Nic30/hdlConvertorAst',
    author='Michal Orsak',
    author_email='Nic30original@gmail.com',
    keywords=['hdl', 'vhdl', 'verilog', 'systemverilog',
              'parser', 'preprocessor', 'antlr4', 'ast', 'code-generator'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
    ],
    license="MIT",
    packages=find_packages(exclude=["tests", ]),
    test_suite='tests.all.suite',
    install_requires=deps,
    test_require=deps,
)




from hdlConvertorAst.hdlAst import HdlTypeBitsDef, HdlOp, HdlOpType, HdlValueId
from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlTypeAuto
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.to.verilog.utils import collect_array_dims, get_wire_t_params
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index, \
    hdl_sub_int


class VerilogResolveTypes(HdlAstVisitor):
    """
    Translate Verilog HDL types to BasicHdlSimModel HDL types
    """
    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        o.type = self.visit_type(o.type)
        if o.value is not None:
            self.visit_iHdlExpr(o.value)

    def visit_type(self, t):
        """
        :type t: iHdlExpr
        """
        t, array_dims = collect_array_dims(t)
        wire_params = get_wire_t_params(t)
        if wire_params is None:
            if t == HdlTypeAuto:
                t = HdlTypeBitsDef(1)
        else:
            base_t, msb, is_signed, _ = wire_params
            lsb = 0
            if msb is None:
                msb = 0
            elif isinstance(msb, HdlOp):
                if msb.fn == HdlOpType.DOWNTO:
                    msb, lsb = msb.ops
                elif msb.fn == HdlOpType.TO:
                    lsb, msb = msb.ops
                elif msb.fn == HdlOpType.CALL and msb.ops[0] == HdlValueId("slice"):
                    lsb = msb.ops[2]
                    msb = hdl_sub_int(msb.ops[1], 1)

            t = HdlTypeBitsDef(msb, lsb=lsb, signed=is_signed)

        for i in array_dims:
            t = hdl_index(t, i)

        return t



from hdlConvertorAst.hdlAst import HdlValueId, HdlOpType, HdlValueInt, HdlOp, iHdlExpr
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword

NONE = HdlValueId("None")


def verilog_slice_to_width(width):
    """
    :type width: iHdlExpr
    :return: Union[int, iHdlExpr]
    """
    if isinstance(width, HdlOp):
        if width.fn in (HdlOpType.DOWNTO, HdlOpType.TO):
            if width.fn == HdlOpType.DOWNTO:
                high, low = width.ops
            else:
                low, high = width.ops
            return _verilog_slice_to_width(high, low)

    return width

def _verilog_slice_to_width(high, low):
    """
    :type high: Union[int, iHdlExpr]
    :type low: Union[int, iHdlExpr]
    :return: Union[int, iHdlExpr]
    """
    if isinstance(low, (int, HdlValueInt)) and isinstance(high, (int, HdlValueInt)):
        assert int(low) == 0, low
        return int(high) + 1

    if isinstance(low, (int, HdlValueInt)) and int(low) == 0:
        w = high
    else:
        w = HdlOp(HdlOpType.SUB, [high, low])

    return HdlOp(HdlOpType.ADD, [w, HdlValueInt(1, None, None)])


def BitsT(width, is_signed=False, bits_cls_name="Bits3t"):
    """
    Create an AST expression of Bits type constructor
    (reg/std_logic_vector equivalent for BasicHdlSimModel)

    :type width: iHdlExpr
    """
    width = verilog_slice_to_width(width)

    if isinstance(width, int):
        width = HdlValueInt(width, None, None)

    c = HdlOp(HdlOpType.CALL, [
        HdlValueId(bits_cls_name, obj=LanguageKeyword()),
        width,
        NONE if is_signed is None else HdlValueInt(int(is_signed), None, None)
    ])
    return c


def sensitivityByOp(op):
    """
    Get sensitivity type for operator.

    :type op: HdlOpType
    :return: Tuple[sensitive on rising edge, sensitive to falling edge]
    """

    if op == HdlOpType.RISING:
        return (True, False)
    elif op == HdlOpType.FALLING:
        return (False, True)
    else:
        raise TypeError(op)



from enum import Enum


# [TODO] use full names of standard like 1800_2012
#   because there are modifications and extensions which differs
#   in first number but they have same second number
#   (e.g. 1800-2017 vs 1802-2017)
class Language(Enum):
    VHDL_2002 = "vhdl2002"
    VHDL_2008 = "vhdl2008"
    VHDL = VHDL_2008  # default vhdl version

    VERILOG_1995 = "verilog1995"  # ieee1364_1995
    VERILOG_2001 = "verilog2001"  # ieee1364_2001
    VERILOG_2001_NOCONFIG = "verilog2001_noconfig"  # ieee1364_2001-noconfig
    VERILOG_2005 = "verilog2005"  # ieee1364_2001
    VERILOG = VERILOG_2001  # default verilog version

    SYSTEM_VERILOG_2005 = "sv2005"  # ieee1800_2005
    SYSTEM_VERILOG_2009 = "sv2009"  # ieee1800_2009
    SYSTEM_VERILOG_2012 = "sv2012"  # ieee1800_2012
    SYSTEM_VERILOG_2017 = "sv2017"  # ieee1800_2017
    SYSTEM_VERILOG = SYSTEM_VERILOG_2017  # default verilog version

    # json produced by hdlConvertor.to.json.ToJson
    HDLCONVERTOR_JSON = "hdlConvertor_JSON"
    HWT = "hwt"

    def is_vhdl(self):
        return self in [Language.VHDL_2002,
                        Language.VHDL_2008]

    def is_verilog(self):
        return self in (
            Language.VERILOG_1995,
            Language.VERILOG_2001,
            Language.VERILOG_2001_NOCONFIG,
            Language.VERILOG_2005
        )

    def is_system_verilog(self):
        return self in (
            Language.SYSTEM_VERILOG_2005,
            Language.SYSTEM_VERILOG_2009,
            Language.SYSTEM_VERILOG_2012,
            Language.SYSTEM_VERILOG_2017
        )



from hdlConvertorAst.hdlAst._expr import HdlOp, HdlOpType, HdlValueId
from hdlConvertorAst.hdlAst import HdlValueInt
from builtins import isinstance


def to_property_call(o, prop_name):
    """
    :note: a * b -> a.prop_name(b)
    :type o: HdlOp
    """
    o.fn = HdlOpType.CALL
    o.ops[0] = hdl_getattr(o.ops[0], prop_name)
    return o


def hdl_name_prefix(prefix_name, o):
    """
    :type o: HdlValueId
    :type prefix_name: iHdlExpr
    :return: HdlOp
    """
    assert isinstance(o, HdlValueId), o
    return HdlOp(HdlOpType.DOT, [prefix_name, o])


def hdl_getattr(o, prop_name):
    """
    :type o: iHdlExpr
    :type prop_name: str
    :return: HdlOp
    """
    return HdlOp(HdlOpType.DOT, [o, HdlValueId(prop_name)])


def hdl_call(o, args):
    """
    :type o: iHdlExpr
    :type args: List[iHdlExpr]
    :return: HdlOp
    """
    return HdlOp(HdlOpType.CALL, [o, ] + args)


def hdl_index(o, i):
    """
    :type o: iHdlExpr
    :type args: List[iHdlExpr]
    :return: HdlOp
    """
    return HdlOp(HdlOpType.INDEX, [o, i])


def hdl_or(*args):
    first = True
    res = None
    for o in args:
        if first:
            res = o
            first = False
        else:
            res = HdlOp(HdlOpType.OR, [res, o])
    assert res is not None
    return res


def hdl_downto(msb, lsb):
    return HdlOp(HdlOpType.DOWNTO, [msb, lsb])


def hdl_map_asoc(o1, o2):
    """
    :type o1: iHdlExpr
    :type o2: iHdlExpr
    :return: HdlOp
    """
    return HdlOp(HdlOpType.MAP_ASSOCIATION, [o1, o2])


def hdl_add_int(a, b):
    """
    :type a: iHdlExpr
    :type b: int
    :return: iHdlExpr
    """
    if b == 0:
        return a
    elif isinstance(a, HdlValueInt):
        return HdlValueInt(int(a.val) + b, a.val, a.base)
    elif isinstance(a, HdlOp) and a.fn == HdlOpType.ADD:
        o0, o1 = a.ops
        if isinstance(o1, HdlValueInt):
            return hdl_add_int(o0, int(o1) + b)
    elif isinstance(a, HdlOp) and a.fn == HdlOpType.SUB:
        o0, o1 = a.ops
        if isinstance(o1, HdlValueInt):
            return hdl_sub_int(o0, int(o1) - b)

    return HdlOp(HdlOpType.ADD, [a, HdlValueInt(b, None, None)])


def hdl_sub_int(a, b):
    """
    :type a: iHdlExpr
    :type b: int
    :return: iHdlExpr
    """
    if b == 0:
        return a
    elif isinstance(a, HdlValueInt):
        return HdlValueInt(int(a.val) - b, a.val, a.base)
    elif isinstance(a, HdlOp) and a.fn == HdlOpType.ADD:
        o0, o1 = a.ops
        if isinstance(o1, HdlValueInt):
            return hdl_add_int(o0, int(o1) - b)
    elif isinstance(a, HdlOp) and a.fn == HdlOpType.SUB:
        o0, o1 = a.ops
        if isinstance(o1, HdlValueInt):
            return hdl_sub_int(o0, int(o1) + b)

    return HdlOp(HdlOpType.SUB, [a, HdlValueInt(b, None, None)])

